"""
unit tests for the mysql_cache cache
"""

import logging

import pytest

import salt.cache.mysql_cache as mysql_cache
from salt.exceptions import SaltCacheError
from tests.support.mock import MagicMock, call, patch

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skipif(
        mysql_cache.MySQLdb is None, reason="No python mysql client installed."
    )
]


@pytest.fixture
def configure_loader_modules():
    return {mysql_cache: {}}


def test_run_query():
    """
    Tests that a SaltCacheError is raised when there is a problem writing to the
    cache file.
    """
    with patch("MySQLdb.connect", MagicMock()) as mock_connect:
        expected_calls = call.cursor().execute("SELECT 1;")
        mysql_cache.run_query(conn=mock_connect, query="SELECT 1;")
        mock_connect.assert_has_calls((expected_calls,), True)


def test_store():
    """
    Tests that the store function writes the data to the serializer for storage.
    """

    mock_connect_client = MagicMock()
    with patch.object(mysql_cache, "_init_client") as mock_init_client:
        with patch.dict(
            mysql_cache.__context__,
            {
                "mysql_table_name": "salt",
                "mysql_client": mock_connect_client,
            },
        ):
            with patch.object(mysql_cache, "run_query") as mock_run_query:
                mock_run_query.return_value = (MagicMock(), 1)

                expected_calls = [
                    call(
                        mock_connect_client,
                        "REPLACE INTO salt (bank, etcd_key, data) values(%s,%s,%s)",
                        args=("minions/minion", "key1", b"\xa4data"),
                    )
                ]

                try:
                    mysql_cache.store(bank="minions/minion", key="key1", data="data")
                except SaltCacheError:
                    pytest.fail("This test should not raise an exception")
                mock_run_query.assert_has_calls(expected_calls, True)

            with patch.object(mysql_cache, "run_query") as mock_run_query:
                mock_run_query.return_value = (MagicMock(), 2)

                expected_calls = [
                    call(
                        mock_connect_client,
                        "REPLACE INTO salt (bank, etcd_key, data) values(%s,%s,%s)",
                        args=("minions/minion", "key2", b"\xa4data"),
                    )
                ]

                try:
                    mysql_cache.store(bank="minions/minion", key="key2", data="data")
                except SaltCacheError:
                    pytest.fail("This test should not raise an exception")
                mock_run_query.assert_has_calls(expected_calls, True)

            with patch.object(mysql_cache, "run_query") as mock_run_query:
                mock_run_query.return_value = (MagicMock(), 0)
                with pytest.raises(SaltCacheError) as exc_info:
                    mysql_cache.store(bank="minions/minion", key="data", data="data")
                expected = "Error storing minions/minion data returned 0"
                assert expected in str(exc_info.value)


def test_fetch():
    """
    Tests that the fetch function reads the data from the serializer for storage.
    """

    with patch.object(mysql_cache, "_init_client") as mock_init_client:
        with patch("MySQLdb.connect") as mock_connect:
            mock_connection = mock_connect.return_value
            cursor = mock_connection.cursor.return_value
            cursor.fetchone.return_value = (b"\xa5hello",)

            with patch.dict(
                mysql_cache.__context__,
                {
                    "mysql_client": mock_connection,
                    "mysql_table_name": "salt",
                },
            ):
                ret = mysql_cache.fetch(bank="bank", key="key")
                assert ret == "hello"


def test_flush():
    """
    Tests the flush function in mysql_cache.
    """
    mock_connect_client = MagicMock()
    with patch.object(mysql_cache, "_init_client") as mock_init_client:
        with patch.dict(
            mysql_cache.__context__,
            {"mysql_client": mock_connect_client, "mysql_table_name": "salt"},
        ):
            with patch.object(mysql_cache, "run_query") as mock_run_query:

                expected_calls = [
                    call(
                        mock_connect_client,
                        "DELETE FROM salt WHERE bank=%s",
                        args=("bank",),
                    ),
                ]
                mock_run_query.return_value = (MagicMock(), "")
                mysql_cache.flush(bank="bank")
                mock_run_query.assert_has_calls(expected_calls, True)

                expected_calls = [
                    call(
                        mock_connect_client,
                        "DELETE FROM salt WHERE bank=%s AND etcd_key=%s",
                        args=("bank", "key"),
                    )
                ]
                mysql_cache.flush(bank="bank", key="key")
                mock_run_query.assert_has_calls(expected_calls, True)


def test_init_client():
    """
    Tests that the _init_client places the correct information in __context__
    """
    with patch.dict(
        mysql_cache.__opts__,
        {"mysql.max_allowed_packet": 100000},
    ):
        with patch.object(mysql_cache, "_create_table") as mock_create_table:
            mysql_cache._init_client()

            assert "mysql_table_name" in mysql_cache.__context__
            assert mysql_cache.__context__["mysql_table_name"] == "salt"

            assert "mysql_kwargs" in mysql_cache.__context__
            assert mysql_cache.__context__["mysql_kwargs"]["autocommit"]
            assert mysql_cache.__context__["mysql_kwargs"]["host"] == "127.0.0.1"
            assert mysql_cache.__context__["mysql_kwargs"]["db"] == "salt_cache"
            assert mysql_cache.__context__["mysql_kwargs"]["port"] == 3306
            assert (
                mysql_cache.__context__["mysql_kwargs"]["max_allowed_packet"] == 100000
            )
            assert not mysql_cache.__context__["mysql_fresh_connection"]

    with patch.dict(
        mysql_cache.__opts__,
        {
            "mysql.max_allowed_packet": 100000,
            "mysql.db": "salt_mysql_db",
            "mysql.host": "mysql-host",
            "mysql.fresh_connection": True,
        },
    ):
        with patch.object(mysql_cache, "_create_table") as mock_create_table:
            mysql_cache._init_client()

            assert "mysql_table_name" in mysql_cache.__context__
            assert mysql_cache.__context__["mysql_table_name"] == "salt"

            assert "mysql_kwargs" in mysql_cache.__context__
            assert mysql_cache.__context__["mysql_kwargs"]["autocommit"]
            assert mysql_cache.__context__["mysql_kwargs"]["host"] == "mysql-host"
            assert mysql_cache.__context__["mysql_kwargs"]["db"] == "salt_mysql_db"
            assert mysql_cache.__context__["mysql_kwargs"]["port"] == 3306
            assert (
                mysql_cache.__context__["mysql_kwargs"]["max_allowed_packet"] == 100000
            )
            assert mysql_cache.__context__["mysql_fresh_connection"]


def test_create_table():
    """
    Tests that the _create_table
    """

    mock_connect_client = MagicMock()
    with patch.dict(
        mysql_cache.__context__,
        {
            "mysql_table_name": "salt",
            "mysql_client": mock_connect_client,
            "mysql_kwargs": {"db": "salt_cache"},
        },
    ):
        with patch.object(mysql_cache, "run_query") as mock_run_query:
            mock_run_query.return_value = (MagicMock(), 1)

            sql_call = """CREATE TABLE IF NOT EXISTS salt (
      bank CHAR(255),
      etcd_key CHAR(255),
      data MEDIUMBLOB,
      last_update TIMESTAMP NOT NULL
                  DEFAULT CURRENT_TIMESTAMP
                  ON UPDATE CURRENT_TIMESTAMP,
      PRIMARY KEY(bank, etcd_key)
    );"""
            expected_calls = [call(mock_connect_client, sql_call)]
            try:
                mysql_cache._create_table()
            except SaltCacheError:
                pytest.fail("This test should not raise an exception")
            mock_run_query.assert_has_calls(expected_calls, True)



import time
from collections import namedtuple

import psutil
import pytest

import salt.modules.ps
import salt.modules.ps as ps
import salt.utils.data
import salt.utils.platform
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mock import MagicMock, Mock, call, patch

# TestCase Exceptions are tested in tests/unit/modules/test_ps.py


@pytest.fixture
def configure_loader_modules():
    return {ps: {}}


@pytest.fixture
def sample_process():
    status = b"fnord"
    extra_data = {
        "utime": "42",
        "stime": "42",
        "children_utime": "42",
        "children_stime": "42",
        "ttynr": "42",
        "cpu_time": "42",
        "blkio_ticks": "99",
        "ppid": "99",
        "cpu_num": "9999999",
    }
    important_data = {
        "name": b"blerp",
        "status": status,
        "create_time": "393829200",
    }
    important_data.update(extra_data)
    patch_stat_file = patch(
        "psutil._psplatform.Process._parse_stat_file",
        return_value=important_data,
        create=True,
    )
    patch_exe = patch(
        "psutil._psplatform.Process.exe",
        return_value=important_data["name"].decode(),
        create=True,
    )
    patch_oneshot = patch(
        "psutil._psplatform.Process.oneshot",
        return_value={
            # These keys can be found in psutil/_psbsd.py
            1: important_data["status"].decode(),
            # create
            9: float(important_data["create_time"]),
            # user
            14: float(important_data["create_time"]),
            # sys
            15: float(important_data["create_time"]),
            # ch_user
            16: float(important_data["create_time"]),
            # ch_sys -- we don't really care what they are, obviously
            17: float(important_data["create_time"]),
            24: important_data["name"].decode(),
        },
        create=True,
    )
    patch_kinfo = patch(
        "psutil._psplatform.Process._get_kinfo_proc",
        return_value={
            # These keys can be found in psutil/_psosx.py
            9: important_data["status"].decode(),
            8: float(important_data["create_time"]),
            10: important_data["name"].decode(),
        },
        create=True,
    )
    patch_status = patch(
        "psutil._psplatform.Process.status", return_value=status.decode()
    )
    patch_create_time = patch(
        "psutil._psplatform.Process.create_time", return_value=393829200
    )
    with (
        patch_stat_file
    ), patch_status, patch_create_time, patch_exe, patch_oneshot, patch_kinfo:
        proc = psutil.Process(pid=42)
        proc.info = proc.as_dict(("name", "status"))
        yield proc


def test__status_when_process_is_found_with_matching_status_then_proc_info_should_be_returned(
    sample_process,
):
    expected_result = [{"pid": 42, "name": "blerp"}]
    proc = sample_process
    with patch(
        "psutil.process_iter",
        autospec=True,
        return_value=[
            proc
        ],  # MagicMock(info={"status": "fnord", "blerp": "whatever"})],
    ):

        actual_result = salt.modules.ps.status(status="fnord")
        assert actual_result == expected_result


def test__status_when_no_matching_processes_then_no_results_should_be_returned():
    expected_result = []
    with patch(
        "psutil.process_iter",
        autospec=True,
        return_value=[MagicMock(info={"status": "foo", "blerp": "whatever"})],
    ):

        actual_result = salt.modules.ps.status(status="fnord")
        assert actual_result == expected_result


def test__status_when_some_matching_processes_then_only_correct_info_should_be_returned(
    sample_process,
):
    expected_result = [{"name": "blerp", "pid": 42}]
    with patch(
        "psutil.process_iter",
        autospec=True,
        return_value=[
            sample_process,
            MagicMock(info={"status": "foo", "name": "wherever", "pid": 9998}),
            MagicMock(info={"status": "bar", "name": "whenever", "pid": 9997}),
        ],
    ):

        actual_result = salt.modules.ps.status(status="fnord")
        assert actual_result == expected_result


STUB_CPU_TIMES = namedtuple("cputimes", "user nice system idle")(1, 2, 3, 4)
STUB_CPU_TIMES_PERCPU = [
    namedtuple("cputimes", "user nice system idle")(1, 2, 3, 4),
    namedtuple("cputimes", "user nice system idle")(1, 2, 3, 4),
    namedtuple("cputimes", "user nice system idle")(1, 2, 3, 4),
    namedtuple("cputimes", "user nice system idle")(1, 2, 3, 4),
]
STUB_VIRT_MEM = namedtuple("vmem", "total available percent used free")(
    1000, 500, 50, 500, 500
)
STUB_SWAP_MEM = namedtuple("swap", "total used free percent sin sout")(
    1000, 500, 500, 50, 0, 0
)
STUB_PHY_MEM_USAGE = namedtuple("usage", "total used free percent")(1000, 500, 500, 50)
STUB_DISK_PARTITION = namedtuple("partition", "device mountpoint fstype, opts")(
    "/dev/disk0s2", "/", "hfs", "rw,local,rootfs,dovolfs,journaled,multilabel"
)
STUB_DISK_USAGE = namedtuple("usage", "total used free percent")(1000, 500, 500, 50)
STUB_NETWORK_IO = namedtuple(
    "iostat",
    "bytes_sent, bytes_recv, packets_sent, packets_recv, errin errout dropin dropout",
)(1000, 2000, 500, 600, 1, 2, 3, 4)
STUB_NETWORK_IO_PERNIC = {
    "lo": STUB_NETWORK_IO,
    "eth0": STUB_NETWORK_IO,
    "eth1": STUB_NETWORK_IO,
}
STUB_DISK_IO = namedtuple(
    "iostat", "read_count, write_count, read_bytes, write_bytes, read_time, write_time"
)(1000, 2000, 500, 600, 2000, 3000)
STUB_DISK_IO_PERDISK = {
    "nvme0n1": STUB_DISK_IO,
    "nvme0n1p1": STUB_DISK_IO,
    "nvme0n1p2": STUB_DISK_IO,
    "nvme0n1p3": STUB_DISK_IO,
}


@pytest.fixture
def stub_memory_usage():
    return namedtuple(
        "vmem",
        "total available percent used free active inactive buffers cached shared",
    )(
        15722012672,
        9329594368,
        40.7,
        5137018880,
        4678086656,
        6991405056,
        2078953472,
        1156378624,
        4750528512,
        898908160,
    )


@pytest.fixture(scope="module")
def stub_user():
    return namedtuple("user", "name, terminal, host, started")(
        "bdobbs", "ttys000", "localhost", 0.0
    )


STUB_PID_LIST = [0, 1, 2, 3]

try:
    import utmp  # pylint: disable=W0611

    HAS_UTMP = True
except ImportError:
    HAS_UTMP = False
# pylint: enable=import-error,unused-import


def _get_proc_name(proc):
    return proc.name()


def _get_proc_pid(proc):
    return proc.pid


class DummyProcess:
    """
    Dummy class to emulate psutil.Process. This ensures that _any_ string
    values used for any of the options passed in are converted to str types on
    both Python 2 and Python 3.
    """

    def __init__(
        self,
        cmdline=None,
        create_time=None,
        name=None,
        status=None,
        username=None,
        pid=None,
        cpu_times=None,
    ):
        self._cmdline = salt.utils.data.decode(
            cmdline if cmdline is not None else [], to_str=True
        )
        self._create_time = salt.utils.data.decode(
            create_time if create_time is not None else time.time(), to_str=True
        )
        self._name = salt.utils.data.decode(
            name if name is not None else [], to_str=True
        )
        self._status = salt.utils.data.decode(status, to_str=True)
        self._username = salt.utils.data.decode(username, to_str=True)
        self._pid = salt.utils.data.decode(
            pid if pid is not None else 12345, to_str=True
        )

        if salt.utils.platform.is_windows():
            scputimes = namedtuple(
                "scputimes", ["user", "system", "children_user", "children_system"]
            )
            dummy_cpu_times = scputimes(7713.79, 1278.44, 17114.2, 2023.36)
        else:
            scputimes = namedtuple(
                "scputimes",
                ["user", "system", "children_user", "children_system", "iowait"],
            )
            dummy_cpu_times = scputimes(7713.79, 1278.44, 17114.2, 2023.36, 0.0)
        self._cpu_times = cpu_times if cpu_times is not None else dummy_cpu_times

    def __enter__(self):
        pass

    def __exit__(self):
        pass

    def cmdline(self):
        return self._cmdline

    def create_time(self):
        return self._create_time

    def name(self):
        return self._name

    def status(self):
        return self._status

    def username(self):
        return self._username

    def pid(self):
        return self._pid

    def cpu_times(self):
        return self._cpu_times


@pytest.fixture
def mocked_proc():
    mocked_proc = MagicMock("psutil.Process")
    mocked_proc.name = Mock(return_value="test_mock_proc")
    mocked_proc.pid = Mock(return_value=9999999999)
    mocked_proc.cmdline = Mock(
        return_value=["test_mock_proc", "--arg", "--kwarg=value"]
    )

    with patch("psutil.Process.send_signal"), patch(
        "psutil.process_iter",
        MagicMock(return_value=[mocked_proc]),
    ):
        yield mocked_proc


def test__get_proc_cmdline():
    cmdline = ["echo", "питон"]
    ret = ps._get_proc_cmdline(DummyProcess(cmdline=cmdline))
    assert ret == cmdline, ret

    with patch.object(DummyProcess, "cmdline") as mock_cmdline:
        mock_cmdline.side_effect = psutil.NoSuchProcess(DummyProcess(cmdline=cmdline))
        ret = ps._get_proc_cmdline(DummyProcess(cmdline=cmdline))
        assert ret == []

    with patch.object(DummyProcess, "cmdline") as mock_cmdline:
        mock_cmdline.side_effect = psutil.AccessDenied(DummyProcess(cmdline=cmdline))
        ret = ps._get_proc_cmdline(DummyProcess(cmdline=cmdline))
        assert ret == []


def test__get_proc_create_time():
    cmdline = ["echo", "питон"]
    create_time = 1694729500.1093624
    ret = ps._get_proc_create_time(
        DummyProcess(cmdline=cmdline, create_time=create_time)
    )
    assert ret == create_time

    with patch.object(DummyProcess, "create_time") as mock_create_time:
        mock_create_time.side_effect = psutil.NoSuchProcess(
            DummyProcess(cmdline=cmdline, create_time=create_time)
        )
        ret = ps._get_proc_create_time(
            DummyProcess(cmdline=cmdline, create_time=create_time)
        )
        assert ret is None

    with patch.object(DummyProcess, "create_time") as mock_create_time:
        mock_create_time.side_effect = psutil.AccessDenied(
            DummyProcess(cmdline=cmdline, create_time=create_time)
        )
        ret = ps._get_proc_create_time(
            DummyProcess(cmdline=cmdline, create_time=create_time)
        )
        assert ret is None


def test__get_proc_name():
    cmdline = ["echo", "питон"]
    proc_name = "proc_name"
    ret = ps._get_proc_name(DummyProcess(cmdline=cmdline, name=proc_name))
    assert ret == proc_name

    with patch.object(DummyProcess, "name") as mock_name:
        mock_name.side_effect = psutil.NoSuchProcess(
            DummyProcess(cmdline=cmdline, name=proc_name)
        )
        ret = ps._get_proc_name(DummyProcess(cmdline=cmdline, name=proc_name))
        assert ret == []

    with patch.object(DummyProcess, "name") as mock_name:
        mock_name.side_effect = psutil.AccessDenied(
            DummyProcess(cmdline=cmdline, name=proc_name)
        )
        ret = ps._get_proc_name(DummyProcess(cmdline=cmdline, name=proc_name))
        assert ret == []


def test__get_proc_status():
    cmdline = ["echo", "питон"]
    proc_status = "sleeping"
    ret = ps._get_proc_status(DummyProcess(cmdline=cmdline, status=proc_status))
    assert ret == proc_status

    with patch.object(DummyProcess, "status") as mock_status:
        mock_status.side_effect = psutil.NoSuchProcess(
            DummyProcess(cmdline=cmdline, status=proc_status)
        )
        ret = ps._get_proc_status(DummyProcess(cmdline=cmdline, status=proc_status))
        assert ret is None

    with patch.object(DummyProcess, "status") as mock_status:
        mock_status.side_effect = psutil.AccessDenied(
            DummyProcess(cmdline=cmdline, status=proc_status)
        )
        ret = ps._get_proc_status(DummyProcess(cmdline=cmdline, status=proc_status))
        assert ret is None


def test__get_proc_username():
    cmdline = ["echo", "питон"]
    proc_username = "root"
    ret = ps._get_proc_username(DummyProcess(cmdline=cmdline, username=proc_username))
    assert ret == proc_username

    with patch.object(DummyProcess, "username") as mock_username:
        mock_username.side_effect = psutil.NoSuchProcess(
            DummyProcess(cmdline=cmdline, username=proc_username)
        )
        ret = ps._get_proc_username(
            DummyProcess(cmdline=cmdline, username=proc_username)
        )
        assert ret is None

    with patch.object(DummyProcess, "username") as mock_username:
        mock_username.side_effect = psutil.AccessDenied(
            DummyProcess(cmdline=cmdline, username=proc_username)
        )
        ret = ps._get_proc_username(
            DummyProcess(cmdline=cmdline, username=proc_username)
        )
        assert ret is None


def test_get_pid_list():
    with patch("psutil.pids", MagicMock(return_value=STUB_PID_LIST)):
        assert STUB_PID_LIST == ps.get_pid_list()


def test_kill_pid():
    cmdline = ["echo", "питон"]
    top_proc = DummyProcess(cmdline=cmdline)

    with patch("psutil.Process") as mock_process:
        mock_process.side_effect = psutil.NoSuchProcess(top_proc)
        ret = ps.kill_pid(0, signal=999)
        assert not ret

    with patch("psutil.Process") as send_signal_mock:
        ps.kill_pid(0, signal=999)
        assert send_signal_mock.call_args == call(0)


def test_pkill(mocked_proc):
    mocked_proc.send_signal = MagicMock()
    test_signal = 1234
    ps.pkill(_get_proc_name(mocked_proc), signal=test_signal)
    assert mocked_proc.send_signal.call_args == call(test_signal)

    mocked_proc.send_signal = MagicMock(side_effect=psutil.NoSuchProcess(mocked_proc))
    ret = ps.pkill(_get_proc_name(mocked_proc), signal=test_signal)
    assert ret is None

    mocked_proc.username = MagicMock(return_value="root")
    with patch.object(ps, "_get_proc_username", return_value=None):
        ret = ps.pkill(_get_proc_name(mocked_proc), signal=test_signal, user="root")
    assert ret is None

    mocked_proc.username = MagicMock(return_value="root")
    ret = ps.pkill(_get_proc_name(mocked_proc), signal=test_signal, user="root")
    assert mocked_proc.send_signal.call_args == call(test_signal)


def test_pgrep(mocked_proc):
    with patch(
        "psutil.process_iter",
        MagicMock(return_value=[mocked_proc]),
    ):
        assert mocked_proc.pid in (ps.pgrep(_get_proc_name(mocked_proc)) or [])

        assert mocked_proc.pid in (
            ps.pgrep(_get_proc_name(mocked_proc), full=True) or []
        )


def test_pgrep_regex(mocked_proc):
    with patch(
        "psutil.process_iter",
        MagicMock(return_value=[mocked_proc]),
    ):
        assert mocked_proc.pid in (
            ps.pgrep("t.st_[a-z]+_proc", pattern_is_regex=True) or []
        )


def test_cpu_percent():
    with patch("psutil.cpu_percent", MagicMock(return_value=1)):
        assert ps.cpu_percent() == 1

    with patch("psutil.cpu_percent", MagicMock(return_value=(1, 1, 1, 1))):
        assert ps.cpu_percent(per_cpu=True) == [1, 1, 1, 1]

    with patch("psutil.cpu_percent", MagicMock(return_value=1)):
        assert ps.cpu_percent(per_cpu=False) == 1


def test_cpu_times():
    with patch("psutil.cpu_times", MagicMock(return_value=STUB_CPU_TIMES)):
        assert {"idle": 4, "nice": 2, "system": 3, "user": 1} == ps.cpu_times()

    with patch(
        "psutil.cpu_times",
        MagicMock(return_value=STUB_CPU_TIMES_PERCPU),
    ):
        assert [
            {"idle": 4, "nice": 2, "system": 3, "user": 1},
            {"idle": 4, "nice": 2, "system": 3, "user": 1},
            {"idle": 4, "nice": 2, "system": 3, "user": 1},
            {"idle": 4, "nice": 2, "system": 3, "user": 1},
        ] == ps.cpu_times(per_cpu=True)


def test_virtual_memory():
    with patch(
        "psutil.virtual_memory",
        MagicMock(return_value=STUB_VIRT_MEM),
    ):
        assert {
            "used": 500,
            "total": 1000,
            "available": 500,
            "percent": 50,
            "free": 500,
        } == ps.virtual_memory()


def test_swap_memory():
    with patch(
        "psutil.swap_memory",
        MagicMock(return_value=STUB_SWAP_MEM),
    ):
        assert {
            "used": 500,
            "total": 1000,
            "percent": 50,
            "free": 500,
            "sin": 0,
            "sout": 0,
        } == ps.swap_memory()


def test_disk_partitions():
    with patch(
        "psutil.disk_partitions",
        MagicMock(return_value=[STUB_DISK_PARTITION]),
    ):
        assert {
            "device": "/dev/disk0s2",
            "mountpoint": "/",
            "opts": "rw,local,rootfs,dovolfs,journaled,multilabel",
            "fstype": "hfs",
        } == ps.disk_partitions()[0]


def test_disk_usage():
    with patch(
        "psutil.disk_usage",
        MagicMock(return_value=STUB_DISK_USAGE),
    ):
        assert {
            "used": 500,
            "total": 1000,
            "percent": 50,
            "free": 500,
        } == ps.disk_usage("DUMMY_PATH")


def test_disk_partition_usage():
    with patch(
        "psutil.disk_partitions",
        MagicMock(return_value=[STUB_DISK_PARTITION]),
    ):
        with patch(
            "psutil.disk_usage",
            MagicMock(return_value=STUB_DISK_USAGE),
        ):
            result = ps.disk_partition_usage()[0]
            assert {
                "device": "/dev/disk0s2",
                "mountpoint": "/",
                "fstype": "hfs",
                "opts": "rw,local,rootfs,dovolfs,journaled,multilabel",
                "total": 1000,
                "used": 500,
                "free": 500,
                "percent": 50,
            } == result


def test_network_io_counters():
    with patch(
        "psutil.net_io_counters",
        MagicMock(return_value=STUB_NETWORK_IO),
    ):
        assert {
            "packets_sent": 500,
            "packets_recv": 600,
            "bytes_recv": 2000,
            "dropout": 4,
            "bytes_sent": 1000,
            "errout": 2,
            "errin": 1,
            "dropin": 3,
        } == ps.network_io_counters()

    with patch(
        "psutil.net_io_counters",
        MagicMock(return_value=STUB_NETWORK_IO_PERNIC),
    ):
        assert {
            "packets_sent": 500,
            "packets_recv": 600,
            "bytes_recv": 2000,
            "dropout": 4,
            "bytes_sent": 1000,
            "errout": 2,
            "errin": 1,
            "dropin": 3,
        } == ps.network_io_counters(interface="eth0")

        assert not ps.network_io_counters(interface="eth2")


def test_disk_io_counters():
    with patch(
        "psutil.disk_io_counters",
        MagicMock(return_value=STUB_DISK_IO),
    ):
        assert {
            "read_time": 2000,
            "write_bytes": 600,
            "read_bytes": 500,
            "write_time": 3000,
            "read_count": 1000,
            "write_count": 2000,
        } == ps.disk_io_counters()

    with patch(
        "psutil.disk_io_counters",
        MagicMock(return_value=STUB_DISK_IO_PERDISK),
    ):
        assert {
            "read_time": 2000,
            "write_bytes": 600,
            "read_bytes": 500,
            "write_time": 3000,
            "read_count": 1000,
            "write_count": 2000,
        } == ps.disk_io_counters(device="nvme0n1p1")

        assert not ps.disk_io_counters(device="nvme0n1p4")


def test_get_users(stub_user):
    with patch("psutil.users", MagicMock(return_value=[stub_user])):
        assert {
            "terminal": "ttys000",
            "started": 0.0,
            "host": "localhost",
            "name": "bdobbs",
        } == ps.get_users()[0]


def test_top():
    """
    See the following issue:

    https://github.com/saltstack/salt/issues/56942
    """
    # Limiting to one process because the test suite might be running as
    # PID 1 under docker and there may only *be* one process running.
    result = ps.top(num_processes=1, interval=0)
    assert len(result) == 1

    cmdline = ["echo", "питон"]
    top_proc = DummyProcess(cmdline=cmdline)

    with patch("psutil.pids", return_value=[1]):
        with patch("psutil.Process") as mock_process:
            mock_process.side_effect = psutil.NoSuchProcess(top_proc)
            ret = ps.top(num_processes=1, interval=0)
            assert ret == []

    if salt.utils.platform.is_windows():
        scputimes = namedtuple(
            "scputimes", ["user", "system", "children_user", "children_system"]
        )
        zombie_cpu_times = scputimes(0, 0, 0, 0)

        smem_info = namedtuple(
            "pmem",
            [
                "rss",
                "vms",
                "num_page_faults",
                "peak_wset",
                "wset",
                "peak_paged_pool",
                "paged_pool",
                "peak_nonpaged_pool",
                "nonpaged_pool28144",
                "pagefile",
                "peak_pagefile",
                "private",
            ],
        )
        zombie_mem_info = smem_info(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
    else:
        scputimes = namedtuple(
            "scputimes",
            ["user", "system", "children_user", "children_system", "iowait"],
        )
        zombie_cpu_times = scputimes(0, 0, 0, 0, 0)

        smem_info = namedtuple(
            "pmem", ["rss", "vms", "shared", "text", "lib", "data", "dirty"]
        )
        zombie_mem_info = smem_info(0, 0, 0, 0, 0, 0, 0)

    with patch("psutil.pids", return_value=[1]):
        with patch("psutil.Process", return_value=top_proc):
            with patch.object(top_proc, "cpu_times") as mock_cpu_times:
                with patch.object(
                    top_proc, "memory_info", return_value=zombie_mem_info, create=True
                ):
                    mock_cpu_times.side_effect = [
                        psutil.ZombieProcess(top_proc),
                        zombie_cpu_times,
                        zombie_cpu_times,
                    ]
                    ret = ps.top(num_processes=1, interval=0)

                    if salt.utils.platform.is_windows():
                        expected_mem = {
                            "rss": 0,
                            "vms": 0,
                            "num_page_faults": 0,
                            "peak_wset": 0,
                            "wset": 0,
                            "peak_paged_pool": 0,
                            "paged_pool": 0,
                            "peak_nonpaged_pool": 0,
                            "nonpaged_pool28144": 0,
                            "pagefile": 0,
                            "peak_pagefile": 0,
                            "private": 0,
                        }

                        expected_cpu = {
                            "user": 0,
                            "system": 0,
                            "children_user": 0,
                            "children_system": 0,
                        }

                    else:
                        expected_mem = {
                            "rss": 0,
                            "vms": 0,
                            "shared": 0,
                            "text": 0,
                            "lib": 0,
                            "data": 0,
                            "dirty": 0,
                        }

                        expected_cpu = {
                            "user": 0,
                            "system": 0,
                            "children_user": 0,
                            "children_system": 0,
                            "iowait": 0,
                        }

                    assert ret[0]["mem"] == expected_mem
                    assert ret[0]["cpu"] == expected_cpu

    with patch("psutil.pids", return_value=[1]):
        with patch("psutil.Process", return_value=top_proc):
            with patch.object(top_proc, "cpu_times") as mock_cpu_times:
                mock_cpu_times.side_effect = [
                    top_proc._cpu_times,
                    psutil.NoSuchProcess(top_proc),
                ]
                ret = ps.top(num_processes=1, interval=0)
                assert ret == []

    with patch("psutil.pids", return_value=[1]):
        with patch("psutil.Process", return_value=top_proc):
            with patch.object(top_proc, "cpu_times") as mock_cpu_times:
                with patch.object(
                    top_proc, "memory_info", create=True
                ) as mock_memory_info:
                    mock_memory_info.side_effect = psutil.NoSuchProcess(top_proc)
                    mock_cpu_times.side_effect = [
                        psutil.ZombieProcess(top_proc),
                        zombie_cpu_times,
                        zombie_cpu_times,
                    ]
                    ret = ps.top(num_processes=1, interval=0)
                    assert ret == []


def test_top_zombie_process():
    # Get 3 pids that are currently running on the system
    pids = psutil.pids()[:3]
    # Get a process instance for each of the pids
    processes = [psutil.Process(pid) for pid in pids]

    # Patch the middle process to raise ZombieProcess when .cpu_times is called
    def raise_exception():
        raise psutil.ZombieProcess(processes[1].pid)

    processes[1].cpu_times = raise_exception

    # Make sure psutil.pids only returns the above 3 pids
    with patch("psutil.pids", return_value=pids):
        # Make sure we use our process list from above
        with patch("psutil.Process", side_effect=processes):
            result = ps.top(num_processes=1, interval=0)
            assert len(result) == 1


def test_status_when_no_status_is_provided_then_raise_invocation_error():
    with pytest.raises(SaltInvocationError):
        actual_result = salt.modules.ps.status(status="")


@pytest.mark.parametrize(
    "exc_type",
    (
        psutil.AccessDenied(pid="9999", name="whatever"),
        psutil.NoSuchProcess(pid="42"),
    ),
)
def test_status_when_access_denied_from_psutil_it_should_CommandExecutionError(
    exc_type,
):
    with patch(
        "psutil.process_iter",
        autospec=True,
        side_effect=exc_type,
    ):
        with pytest.raises(
            salt.exceptions.CommandExecutionError,
            match="Psutil did not return a list of processes",
        ):
            actual_result = salt.modules.ps.status(status="fnord")


def test_status_when_no_filter_is_provided_then_raise_invocation_error():
    with pytest.raises(SaltInvocationError) as invoc_issue:
        actual_result = salt.modules.ps.status(status="")


def test_status_when_access_denied_from_psutil_then_raise_exception():
    with patch(
        "psutil.process_iter",
        autospec=True,
        return_value=psutil.AccessDenied(pid="9999", name="whatever"),
    ):
        with pytest.raises(Exception) as general_issue:
            actual_result = salt.modules.ps.status(status="fnord")


## This is commented out pending discussion on https://github.com/saltstack/salt/commit/2e5c3162ef87cca8a2c7b12ade7c7e1b32028f0a
# @skipIf(not HAS_UTMP, "The utmp module must be installed to run test_get_users_utmp()")
# @patch('psutil.get_users', new=MagicMock(return_value=None))  # This will force the function to use utmp
# def test_get_users_utmp():
#     pass


def test_psaux():
    """
    Testing psaux function in the ps module
    """

    cmd_run_mock = """
USER         PID %CPU %MEM    VSZ   RSS TTY      STAT START   TIME COMMAND
root           1  0.0  0.0 171584 15740 ?        Ss   Aug09   4:18 /usr/lib/systemd/systemd --system --deserialize=83
root           2  0.0  0.0      0     0 ?        S    Aug09   0:02 [kthreadd]
root     2710129  0.0  0.0  18000  7428 pts/4    S+   Aug21   0:33 sudo -E salt-master -l debug
root     2710131  0.0  0.0  18000  1196 pts/6    Ss   Aug21   0:00 sudo -E salt-master -l debug
"""

    with patch.dict(ps.__salt__, {"cmd.run": MagicMock(return_value=cmd_run_mock)}):
        expected = [
            "salt-master",
            [
                "root     2710129  0.0  0.0  18000  7428 pts/4    S+   Aug21   0:33 sudo -E salt-master -l debug",
                "root     2710131  0.0  0.0  18000  1196 pts/6    Ss   Aug21   0:00 sudo -E salt-master -l debug",
            ],
            "2 occurrence(s).",
        ]
        ret = ps.psaux("salt-master")
        assert ret == expected

        expected = ["salt-minion", [], "0 occurrence(s)."]
        ret = ps.psaux("salt-minion")
        assert ret == expected


@pytest.mark.skip_on_windows(reason="ss not available in Windows")
def test_ss():
    """
    Testing ss function in the ps module
    """

    cmd_run_mock = """
tcp   LISTEN     0      128                                                                               0.0.0.0:22                                            0.0.0.0:*        ino:31907 sk:364b cgroup:/system.slice/sshd.service <->

tcp   LISTEN     0      128                                                                                  [::]:22                                               [::]:*        ino:31916 sk:36c4 cgroup:/system.slice/sshd.service v6only:1 <->
"""

    with patch(
        "salt.utils.path.which", MagicMock(return_value="/usr/sbin/ss")
    ), patch.dict(ps.__salt__, {"cmd.run": MagicMock(return_value=cmd_run_mock)}):
        expected = [
            "sshd",
            [
                "tcp   LISTEN     0      128                                                                               0.0.0.0:22                                            0.0.0.0:*        ino:31907 sk:364b cgroup:/system.slice/sshd.service <->",
                "tcp   LISTEN     0      128                                                                                  [::]:22                                               [::]:*        ino:31916 sk:36c4 cgroup:/system.slice/sshd.service v6only:1 <->",
            ],
        ]
        ret = ps.ss("sshd")
        assert ret == expected

        expected = ["apache2", []]
        ret = ps.ss("apache2")
        assert ret == expected


def test_netstat():
    """
    Testing netstat function in the ps module
    """

    cmd_run_mock = """
Active Internet connections (servers and established)
Proto Recv-Q Send-Q Local Address           Foreign Address         State       PID/Program name
tcp        0      0 0.0.0.0:22              0.0.0.0:*               LISTEN      668/sshd: /usr/sbin
tcp6       0      0 :::22                   :::*                    LISTEN      668/sshd: /usr/sbin
"""

    with patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/netstat")):
        with patch.dict(ps.__salt__, {"cmd.run": MagicMock(return_value=cmd_run_mock)}):
            expected = [
                "sshd",
                [
                    "tcp        0      0 0.0.0.0:22              0.0.0.0:*               LISTEN      668/sshd: /usr/sbin",
                    "tcp6       0      0 :::22                   :::*                    LISTEN      668/sshd: /usr/sbin",
                ],
            ]
            ret = ps.netstat("sshd")
            assert ret == expected

            expected = ["apache2", []]
            ret = ps.netstat("apache2")
            assert ret == expected


def test_lsof():
    """
    Testing lsof function in the ps module
    """

    sshd_cmd_run_mock = """
COMMAND  PID USER   FD   TYPE             DEVICE SIZE/OFF    NODE NAME
sshd    1743 root  cwd    DIR              254,0     4096       2 /
sshd    1743 root  rtd    DIR              254,0     4096       2 /
sshd    1743 root  txt    REG              254,0   925000 7533685 /usr/bin/sshd (deleted)
sshd    1743 root  DEL    REG              254,0          7481413 /usr/lib/libc.so.6
sshd    1743 root  DEL    REG              254,0          7477716 /usr/lib/libcrypto.so.3
sshd    1743 root  mem    REG              254,0    26520 7482162 /usr/lib/libcap-ng.so.0.0.0
sshd    1743 root  DEL    REG              254,0          7512187 /usr/lib/libresolv.so.2
sshd    1743 root  mem    REG              254,0    22400 7481786 /usr/lib/libkeyutils.so.1.10
sshd    1743 root  mem    REG              254,0    55352 7480841 /usr/lib/libkrb5support.so.0.1
sshd    1743 root  mem    REG              254,0    18304 7475778 /usr/lib/libcom_err.so.2.1
sshd    1743 root  mem    REG              254,0   182128 7477432 /usr/lib/libk5crypto.so.3.1
sshd    1743 root  DEL    REG              254,0          7485543 /usr/lib/libaudit.so.1.0.0
sshd    1743 root  DEL    REG              254,0          7485432 /usr/lib/libz.so.1.2.13
sshd    1743 root  mem    REG              254,0   882552 7480814 /usr/lib/libkrb5.so.3.3
sshd    1743 root  mem    REG              254,0   344160 7475833 /usr/lib/libgssapi_krb5.so.2.2
sshd    1743 root  mem    REG              254,0    67536 7482132 /usr/lib/libpam.so.0.85.1
sshd    1743 root  mem    REG              254,0   165832 7481746 /usr/lib/libcrypt.so.2.0.0
sshd    1743 root  DEL    REG              254,0          7480993 /usr/lib/ld-linux-x86-64.so.2
sshd    1743 root    0r   CHR                1,3      0t0       4 /dev/null
sshd    1743 root    1u  unix 0x0000000000000000      0t0   32930 type=STREAM (CONNECTED)
sshd    1743 root    2u  unix 0x0000000000000000      0t0   32930 type=STREAM (CONNECTED)
sshd    1743 root    3u  IPv4              31907      0t0     TCP *:ssh (LISTEN)
sshd    1743 root    4u  IPv6              31916      0t0     TCP *:ssh (LISTEN)
"""

    apache2_cmd_run_mock = ""

    with patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/netstat")):
        with patch.dict(
            ps.__salt__, {"cmd.run": MagicMock(return_value=sshd_cmd_run_mock)}
        ):
            expected = [
                "sshd",
                "\nCOMMAND  PID USER   FD   TYPE             DEVICE SIZE/OFF    NODE NAME\nsshd    1743 root  cwd    DIR              254,0     4096       2 /\nsshd    1743 root  rtd    DIR              254,0     4096       2 /\nsshd    1743 root  txt    REG              254,0   925000 7533685 /usr/bin/sshd (deleted)\nsshd    1743 root  DEL    REG              254,0          7481413 /usr/lib/libc.so.6\nsshd    1743 root  DEL    REG              254,0          7477716 /usr/lib/libcrypto.so.3\nsshd    1743 root  mem    REG              254,0    26520 7482162 /usr/lib/libcap-ng.so.0.0.0\nsshd    1743 root  DEL    REG              254,0          7512187 /usr/lib/libresolv.so.2\nsshd    1743 root  mem    REG              254,0    22400 7481786 /usr/lib/libkeyutils.so.1.10\nsshd    1743 root  mem    REG              254,0    55352 7480841 /usr/lib/libkrb5support.so.0.1\nsshd    1743 root  mem    REG              254,0    18304 7475778 /usr/lib/libcom_err.so.2.1\nsshd    1743 root  mem    REG              254,0   182128 7477432 /usr/lib/libk5crypto.so.3.1\nsshd    1743 root  DEL    REG              254,0          7485543 /usr/lib/libaudit.so.1.0.0\nsshd    1743 root  DEL    REG              254,0          7485432 /usr/lib/libz.so.1.2.13\nsshd    1743 root  mem    REG              254,0   882552 7480814 /usr/lib/libkrb5.so.3.3\nsshd    1743 root  mem    REG              254,0   344160 7475833 /usr/lib/libgssapi_krb5.so.2.2\nsshd    1743 root  mem    REG              254,0    67536 7482132 /usr/lib/libpam.so.0.85.1\nsshd    1743 root  mem    REG              254,0   165832 7481746 /usr/lib/libcrypt.so.2.0.0\nsshd    1743 root  DEL    REG              254,0          7480993 /usr/lib/ld-linux-x86-64.so.2\nsshd    1743 root    0r   CHR                1,3      0t0       4 /dev/null\nsshd    1743 root    1u  unix 0x0000000000000000      0t0   32930 type=STREAM (CONNECTED)\nsshd    1743 root    2u  unix 0x0000000000000000      0t0   32930 type=STREAM (CONNECTED)\nsshd    1743 root    3u  IPv4              31907      0t0     TCP *:ssh (LISTEN)\nsshd    1743 root    4u  IPv6              31916      0t0     TCP *:ssh (LISTEN)\n",
            ]
            ret = ps.lsof("sshd")
            assert ret == expected

        with patch.dict(
            ps.__salt__, {"cmd.run": MagicMock(return_value=apache2_cmd_run_mock)}
        ):
            expected = ["apache2", ""]
            ret = ps.lsof("apache2")
            assert ret == expected


def test_boot_time():
    """
    Testing boot_time function in the ps module
    """

    with patch("psutil.boot_time", MagicMock(return_value=1691593290.0)):
        expected = 1691593290
        ret = ps.boot_time()
        assert ret == expected

        expected = "08/09/2023"
        ret = ps.boot_time(time_format="%m/%d/%Y")
        assert ret == expected

    with patch("psutil.boot_time") as mock_boot_time:
        mock_boot_time.side_effect = [AttributeError(), 1691593290.0]
        expected = 1691593290
        ret = ps.boot_time()
        assert ret == expected


def test_num_cpus():
    """
    Testing num_cpus function in the ps module
    """

    with patch("psutil.cpu_count") as mock_cpu_count:
        mock_cpu_count.side_effect = AttributeError()
        with patch("psutil.NUM_CPUS", create=True, new=5):
            ret = ps.num_cpus()
            assert ret == 5

    with patch("psutil.cpu_count") as mock_cpu_count:
        mock_cpu_count.return_value = 5
        ret = ps.num_cpus()
        assert ret == 5


def test_total_physical_memory(stub_memory_usage):
    """
    Testing total_physical_memory function in the ps module
    """
    with patch("psutil.virtual_memory") as mock_total_physical_memory:
        mock_total_physical_memory.side_effect = AttributeError()
        with patch(
            "psutil.TOTAL_PHYMEM",
            create=True,
            new=stub_memory_usage.total,
        ):
            ret = ps.total_physical_memory()
            assert ret == 15722012672

    with patch("psutil.virtual_memory") as mock_total_physical_memory:
        mock_total_physical_memory.return_value = stub_memory_usage
        ret = ps.total_physical_memory()
        assert ret == 15722012672


def test_proc_info():
    """
    Testing proc_info function in the ps module
    """
    status = b"fnord"
    extra_data = {
        "utime": "42",
        "stime": "42",
        "children_utime": "42",
        "children_stime": "42",
        "ttynr": "42",
        "cpu_time": "42",
        "blkio_ticks": "99",
        "ppid": "99",
        "cpu_num": "9999999",
    }
    important_data = {
        "name": b"blerp",
        "status": status,
        "create_time": "393829200",
        "username": "root",
    }
    important_data.update(extra_data)
    status_file = b"Name:\tblerp\nUmask:\t0000\nState:\tI (idle)\nTgid:\t99\nNgid:\t0\nPid:\t99\nPPid:\t2\nTracerPid:\t0\nUid:\t0\t0\t0\t0\nGid:\t0\t0\t0\t0\nFDSize:\t64\nGroups:\t \nNStgid:\t99\nNSpid:\t99\nNSpgid:\t0\nNSsid:\t0\nThreads:\t1\nSigQ:\t3/256078\nSigPnd:\t0000000000000000\nShdPnd:\t0000000000000000\nSigBlk:\t0000000000000000\nSigIgn:\tffffffffffffffff\nSigCgt:\t0000000000000000\nCapInh:\t0000000000000000\nCapPrm:\t000001ffffffffff\nCapEff:\t000001ffffffffff\nCapBnd:\t000001ffffffffff\nCapAmb:\t0000000000000000\nNoNewPrivs:\t0\nSeccomp:\t0\nSeccomp_filters:\t0\nSpeculation_Store_Bypass:\tthread vulnerable\nSpeculationIndirectBranch:\tconditional enabled\nCpus_allowed:\tfff\nCpus_allowed_list:\t0-11\nMems_allowed:\t00000001\nMems_allowed_list:\t0\nvoluntary_ctxt_switches:\t2\nnonvoluntary_ctxt_switches:\t0\n"

    patch_stat_file = patch(
        "psutil._psplatform.Process._parse_stat_file",
        return_value=important_data,
        create=True,
    )
    patch_exe = patch(
        "psutil._psplatform.Process.exe",
        return_value=important_data["name"].decode(),
        create=True,
    )
    patch_oneshot = patch(
        "psutil._psplatform.Process.oneshot",
        return_value={
            # These keys can be found in psutil/_psbsd.py
            1: important_data["status"].decode(),
            # create
            9: float(important_data["create_time"]),
            # user
            14: float(important_data["create_time"]),
            # sys
            15: float(important_data["create_time"]),
            # ch_user
            16: float(important_data["create_time"]),
            # ch_sys -- we don't really care what they are, obviously
            17: float(important_data["create_time"]),
            24: important_data["name"].decode(),
        },
        create=True,
    )
    patch_kinfo = patch(
        "psutil._psplatform.Process._get_kinfo_proc",
        return_value={
            # These keys can be found in psutil/_psosx.py
            9: important_data["status"].decode(),
            8: float(important_data["create_time"]),
            10: important_data["name"].decode(),
        },
        create=True,
    )
    patch_status = patch(
        "psutil._psplatform.Process.status", return_value=status.decode()
    )
    patch_create_time = patch(
        "psutil._psplatform.Process.create_time", return_value=393829200
    )
    with (
        patch_stat_file
    ), patch_status, patch_create_time, patch_exe, patch_oneshot, patch_kinfo:
        if salt.utils.platform.is_windows():
            with patch("psutil._pswindows.cext") as mock__psutil_windows:
                with patch("psutil._pswindows.Process.ppid", return_value=99):
                    mock__psutil_windows.proc_username.return_value = (
                        "NT Authority",
                        "System",
                    )

                    expected = {"ppid": 99, "username": r"NT Authority\System"}
                    actual_result = salt.modules.ps.proc_info(
                        pid=99, attrs=["username", "ppid"]
                    )
                    assert actual_result == expected

                    expected = {"pid": 99, "name": "blerp"}
                    actual_result = salt.modules.ps.proc_info(
                        pid=99, attrs=["pid", "name"]
                    )
                    assert actual_result == expected
        else:
            patch_read_status_file = patch(
                "psutil._psplatform.Process._read_status_file", return_value=status_file
            )
            with patch_read_status_file:
                expected = {"ppid": 99, "username": "root"}
                actual_result = salt.modules.ps.proc_info(
                    pid=99, attrs=["username", "ppid"]
                )
                assert actual_result == expected

                expected = {"pid": 99, "name": "blerp"}
                actual_result = salt.modules.ps.proc_info(pid=99, attrs=["pid", "name"])
                assert actual_result == expected


def test_proc_info_access_denied():
    """
    Testing proc_info function in the ps module
    when an AccessDenied exception occurs
    """
    cmdline = ["echo", "питон"]
    dummy_proc = DummyProcess(cmdline=cmdline)
    with patch("psutil.Process") as mock_process:
        mock_process.side_effect = psutil.AccessDenied(dummy_proc)
        with pytest.raises(CommandExecutionError):
            salt.modules.ps.proc_info(pid=99, attrs=["username", "ppid"])


def test_proc_info_no_such_process():
    """
    Testing proc_info function in the ps module
    when an NoSuchProcess exception occurs
    """
    cmdline = ["echo", "питон"]
    dummy_proc = DummyProcess(cmdline=cmdline)
    with patch("psutil.Process") as mock_process:
        mock_process.side_effect = psutil.NoSuchProcess(dummy_proc)
        with pytest.raises(CommandExecutionError):
            salt.modules.ps.proc_info(pid=99, attrs=["username", "ppid"])


def test_proc_info_attribute_error():
    """
    Testing proc_info function in the ps module
    when an AttributeError exception occurs
    """
    with patch("psutil.Process") as mock_process:
        mock_process.side_effect = AttributeError()
        with pytest.raises(CommandExecutionError):
            salt.modules.ps.proc_info(pid=99, attrs=["username", "ppid"])


def test__virtual__no_psutil():
    """
    Test __virtual__ function
    """
    with patch.object(ps, "HAS_PSUTIL", False):
        expected = (
            False,
            "The ps module cannot be loaded: python module psutil not installed.",
        )
        result = ps.__virtual__()
    assert result == expected



"""
Module for manageing PagerDuty resource

:configuration: This module can be used by specifying the name of a
    configuration profile in the minion config, minion pillar, or master
    config.  The default configuration profile name is 'pagerduty.'

    For example:

    .. code-block:: yaml

        pagerduty:
            pagerduty.api_key: F3Rbyjbve43rfFWf2214
            pagerduty.subdomain: mysubdomain


For PagerDuty API details, see https://developer.pagerduty.com/documentation/rest

"""

import requests

import salt.utils.json


def __virtual__():
    """
    No dependencies outside of what Salt itself requires
    """
    return True


def get_users(profile="pagerduty", subdomain=None, api_key=None):
    """
    List users belonging to this account

    CLI Example:

    .. code-block:: bash

        salt myminion pagerduty.get_users
    """

    return _list_items(
        "users",
        "id",
        profile=profile,
        subdomain=subdomain,
        api_key=api_key,
    )


def get_services(profile="pagerduty", subdomain=None, api_key=None):
    """
    List services belonging to this account

    CLI Example:

    .. code-block:: bash

        salt myminion pagerduty.get_services
    """

    return _list_items(
        "services",
        "id",
        profile=profile,
        subdomain=subdomain,
        api_key=api_key,
    )


def get_schedules(profile="pagerduty", subdomain=None, api_key=None):
    """
    List schedules belonging to this account

    CLI Example:

    .. code-block:: bash

        salt myminion pagerduty.get_schedules
    """

    return _list_items(
        "schedules",
        "id",
        profile=profile,
        subdomain=subdomain,
        api_key=api_key,
    )


def get_escalation_policies(profile="pagerduty", subdomain=None, api_key=None):
    """
    List escalation_policies belonging to this account

    CLI Example:

    .. code-block:: bash

        salt myminion pagerduty.get_escalation_policies
    """

    return _list_items(
        "escalation_policies",
        "id",
        profile=profile,
        subdomain=subdomain,
        api_key=api_key,
    )


def _list_items(action, key, profile=None, subdomain=None, api_key=None):
    """
    List items belonging to an API call.

    This method should be in utils.pagerduty.
    """
    items = _query(profile=profile, subdomain=subdomain, api_key=api_key, action=action)
    ret = {}
    for item in items[action]:
        ret[item[key]] = item
    return ret


def _query(
    method="GET",
    profile=None,
    url=None,
    path="api/v1",
    action=None,
    api_key=None,
    service=None,
    params=None,
    data=None,
    subdomain=None,
    verify_ssl=True,
):
    """
    Query the PagerDuty API.

    This method should be in utils.pagerduty.

    """

    if profile:
        creds = __salt__["config.option"](profile)
    else:
        creds = {
            "pagerduty.api_key": api_key,
            "pagerduty.subdomain": subdomain,
        }

    if url is None:
        url = "https://{}.pagerduty.com/{}/{}".format(
            creds["pagerduty.subdomain"], path, action
        )

    if params is None:
        params = {}

    if data is None:
        data = {}

    headers = {"Authorization": "Token token={}".format(creds["pagerduty.api_key"])}

    if method != "GET":
        headers["Content-type"] = "application/json"

    result = requests.request(
        method,
        url,
        headers=headers,
        params=params,
        data=salt.utils.json.dumps(data),
        verify=verify_ssl,
        timeout=120,
    )

    if result.text is None or result.text == "":
        return None
    result_json = result.json()
    # if this query supports pagination, loop and fetch all results, merge them together
    if "total" in result_json and "offset" in result_json and "limit" in result_json:
        offset = result_json["offset"]
        limit = result_json["limit"]
        total = result_json["total"]
        while offset + limit < total:
            offset = offset + limit
            limit = 100
            data["offset"] = offset
            data["limit"] = limit
            next_page_results = requests.request(
                method,
                url,
                headers=headers,
                params=params,
                data=data,  # Already serialized above, don't do it again
                verify=verify_ssl,
                timeout=120,
            ).json()
            offset = next_page_results["offset"]
            limit = next_page_results["limit"]
            # merge results
            for k, v in result_json.items():
                if isinstance(v, list):
                    result_json[k] += next_page_results[k]
    return result_json


def _get_resource_id(resource):
    """
    helper method to find the resource id, since PD API doesn't always return it in the same way
    """
    if "id" in resource:
        return resource["id"]
    if "schedule" in resource:
        return resource["schedule"]["id"]
    return None


def get_resource(
    resource_name,
    key,
    identifier_fields,
    profile="pagerduty",
    subdomain=None,
    api_key=None,
):
    """
    Get any single pagerduty resource by key.

    We allow flexible lookup by any of a list of identifier_fields.
    So, for example, you can look up users by email address or name by calling:

            get_resource('users', key, ['name', 'email'], ...)

    This method is mainly used to translate state sls into pagerduty id's for dependent objects.
    For example, a pagerduty escalation policy contains one or more schedules, which must be passed
    by their pagerduty id.  We look up the schedules by name (using this method), and then translate
    the names into id's.

    This method is implemented by getting all objects of the resource type (cached into __context__),
    then brute force searching through the list and trying to match any of the identifier_fields.
    The __context__ cache is purged after any create, update or delete to the resource.
    """
    # cache the expensive 'get all resources' calls into __context__ so that we do them once per salt run
    if "pagerduty_util.resource_cache" not in __context__:
        __context__["pagerduty_util.resource_cache"] = {}
    if resource_name not in __context__["pagerduty_util.resource_cache"]:
        if resource_name == "services":
            action = resource_name + "?include[]=escalation_policy"
        else:
            action = resource_name
        __context__["pagerduty_util.resource_cache"][resource_name] = _query(
            action=action, profile=profile, subdomain=subdomain, api_key=api_key
        )[resource_name]
    for resource in __context__["pagerduty_util.resource_cache"][resource_name]:
        for field in identifier_fields:
            if resource[field] == key:
                # PagerDuty's /schedules endpoint returns less data than /schedules/:id.
                # so, now that we found the schedule, we need to get all the data for it.
                if resource_name == "schedules":
                    full_resource_info = _query(
                        action="{}/{}".format(resource_name, resource["id"]),
                        profile=profile,
                        subdomain=subdomain,
                        api_key=api_key,
                    )
                    return full_resource_info
                return resource
    return None


def create_or_update_resource(
    resource_name,
    identifier_fields,
    data,
    diff=None,
    profile="pagerduty",
    subdomain=None,
    api_key=None,
):
    """
    create or update any pagerduty resource
    Helper method for present().

    Determining if two resources are the same is different for different PD resource, so this method accepts a diff function.
    The diff function will be invoked as diff(state_information, object_returned_from_pagerduty), and
    should return a dict of data to pass to the PagerDuty update API method, or None if no update
    is to be performed.  If no diff method is provided, the default behavor is to scan the keys in the state_information,
    comparing the matching values in the object_returned_from_pagerduty, and update any values that differ.

    examples:
        create_or_update_resource("user", ["id","name","email"])
        create_or_update_resource("escalation_policies", ["id","name"], diff=my_diff_function)

    """
    # try to locate the resource by any of the identifier_fields that are specified in data
    resource = None
    for field in identifier_fields:
        if field in data:
            resource = get_resource(
                resource_name,
                data[field],
                identifier_fields,
                profile,
                subdomain,
                api_key,
            )
            if resource is not None:
                break

    if resource is None:
        if __opts__["test"]:
            return "would create"
        # flush the resource_cache, because we're modifying a resource
        del __context__["pagerduty_util.resource_cache"][resource_name]
        # create
        return _query(
            method="POST",
            action=resource_name,
            data=data,
            profile=profile,
            subdomain=subdomain,
            api_key=api_key,
        )
    else:
        # update
        data_to_update = {}
        # if differencing function is provided, use it
        if diff:
            data_to_update = diff(data, resource)
        # else default to naive key-value walk of the dicts
        else:
            for k, v in data.items():
                if k.startswith("_"):
                    continue
                resource_value = resource.get(k, None)
                if resource_value is not None and resource_value != v:
                    data_to_update[k] = v
        if data_to_update:
            if __opts__["test"]:
                return "would update"
            # flush the resource_cache, because we're modifying a resource
            del __context__["pagerduty_util.resource_cache"][resource_name]
            resource_id = _get_resource_id(resource)
            return _query(
                method="PUT",
                action=f"{resource_name}/{resource_id}",
                data=data_to_update,
                profile=profile,
                subdomain=subdomain,
                api_key=api_key,
            )
        else:
            return True


def delete_resource(
    resource_name,
    key,
    identifier_fields,
    profile="pagerduty",
    subdomain=None,
    api_key=None,
):
    """
    delete any pagerduty resource

    Helper method for absent()

    example:
            delete_resource("users", key, ["id","name","email"]) # delete by id or name or email

    """
    resource = get_resource(
        resource_name, key, identifier_fields, profile, subdomain, api_key
    )
    if resource:
        if __opts__["test"]:
            return "would delete"
        # flush the resource_cache, because we're modifying a resource
        del __context__["pagerduty_util.resource_cache"][resource_name]
        resource_id = _get_resource_id(resource)
        return _query(
            method="DELETE",
            action=f"{resource_name}/{resource_id}",
            profile=profile,
            subdomain=subdomain,
            api_key=api_key,
        )
    else:
        return True


def resource_present(
    resource,
    identifier_fields,
    diff=None,
    profile="pagerduty",
    subdomain=None,
    api_key=None,
    **kwargs,
):
    """
    Generic resource.present state method.   Pagerduty state modules should be a thin wrapper over this method,
    with a custom diff function.

    This method calls create_or_update_resource() and formats the result as a salt state return value.

    example:
            resource_present("users", ["id","name","email"])
    """

    ret = {"name": kwargs["name"], "changes": {}, "result": None, "comment": ""}
    result = create_or_update_resource(
        resource,
        identifier_fields,
        kwargs,
        diff=diff,
        profile=profile,
        subdomain=subdomain,
        api_key=api_key,
    )
    if result is True:
        pass
    elif result is None:
        ret["result"] = True
    elif __opts__["test"]:
        ret["comment"] = result
    elif "error" in result:
        ret["result"] = False
        ret["comment"] = result
    else:
        ret["result"] = True
        ret["comment"] = result
    return ret


def resource_absent(
    resource,
    identifier_fields,
    profile="pagerduty",
    subdomain=None,
    api_key=None,
    **kwargs,
):
    """
    Generic resource.absent state method.   Pagerduty state modules should be a thin wrapper over this method,
    with a custom diff function.

    This method calls delete_resource() and formats the result as a salt state return value.

    example:
            resource_absent("users", ["id","name","email"])
    """
    ret = {"name": kwargs["name"], "changes": {}, "result": None, "comment": ""}
    for k, v in kwargs.items():
        if k not in identifier_fields:
            continue
        result = delete_resource(
            resource,
            v,
            identifier_fields,
            profile=profile,
            subdomain=subdomain,
            api_key=api_key,
        )
        if result is None:
            ret["result"] = True
            ret["comment"] = f"{v} deleted"
            return ret
        elif result is True:
            continue
        elif __opts__["test"]:
            ret["comment"] = result
            return ret
        elif "error" in result:
            ret["result"] = False
            ret["comment"] = result
            return ret
    return ret



import logging
import textwrap

import pytest

import salt.modules.cmdmod as cmdmod
import salt.modules.file as filemod
from salt.exceptions import SaltInvocationError
from tests.support.mock import Mock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def _common_patches():
    with patch("salt.utils.platform.is_aix", Mock(return_value=False)), patch(
        "os.path.exists", Mock(return_value=True)
    ), patch("salt.utils.path.which", Mock(return_value="fnord")):
        yield


@pytest.fixture
def configure_loader_modules(_common_patches):
    return {
        filemod: {"__salt__": {"cmd.run": cmdmod.run}},
    }


def test_if_lsattr_is_missing_it_should_return_None():
    patch_which = patch(
        "salt.utils.path.which",
        Mock(return_value=None),
    )
    with patch_which:
        actual = filemod.lsattr("foo")
        assert actual is None, actual


def test_on_aix_lsattr_should_be_None():
    patch_aix = patch(
        "salt.utils.platform.is_aix",
        Mock(return_value=True),
    )
    with patch_aix:
        # SaltInvocationError will be raised if filemod.lsattr
        # doesn't early exit
        actual = filemod.lsattr("foo")
        assert actual is None


def test_SaltInvocationError_should_be_raised_when_file_is_missing():
    patch_exists = patch(
        "os.path.exists",
        Mock(return_value=False),
    )
    with patch_exists, pytest.raises(SaltInvocationError):
        filemod.lsattr("foo")


def test_if_chattr_version_is_less_than_required_flags_should_ignore_extended():
    fname = "/path/to/fnord"
    with_extended = (
        textwrap.dedent(
            """
        aAcCdDeijPsStTu---- {}
        """
        )
        .strip()
        .format(fname)
    )
    expected = set("acdijstuADST")
    patch_has_ext = patch(
        "salt.modules.file._chattr_has_extended_attrs",
        Mock(return_value=False),
    )
    patch_run = patch.dict(
        filemod.__salt__,
        {"cmd.run": Mock(return_value=with_extended)},
    )
    with patch_has_ext, patch_run:
        actual = set(filemod.lsattr(fname)[fname])
        msg = f"Actual: {actual!r} Expected: {expected!r}"
        assert actual == expected, msg


def test_if_chattr_version_is_high_enough_then_extended_flags_should_be_returned():
    fname = "/path/to/fnord"
    with_extended = (
        textwrap.dedent(
            """
        aAcCdDeijPsStTu---- {}
        """
        )
        .strip()
        .format(fname)
    )
    expected = set("aAcCdDeijPsStTu")
    patch_has_ext = patch(
        "salt.modules.file._chattr_has_extended_attrs",
        Mock(return_value=True),
    )
    patch_run = patch.dict(
        filemod.__salt__,
        {"cmd.run": Mock(return_value=with_extended)},
    )
    with patch_has_ext, patch_run:
        actual = set(filemod.lsattr(fname)[fname])
        msg = f"Actual: {actual!r} Expected: {expected!r}"
        assert actual == expected, msg


def test_if_supports_extended_but_there_are_no_flags_then_none_should_be_returned():
    fname = "/path/to/fnord"
    with_extended = (
        textwrap.dedent(
            """
        ------------------- {}
        """
        )
        .strip()
        .format(fname)
    )
    expected = set("")
    patch_has_ext = patch(
        "salt.modules.file._chattr_has_extended_attrs",
        Mock(return_value=True),
    )
    patch_run = patch.dict(
        filemod.__salt__,
        {"cmd.run": Mock(return_value=with_extended)},
    )
    with patch_has_ext, patch_run:
        actual = set(filemod.lsattr(fname)[fname])
        msg = f"Actual: {actual!r} Expected: {expected!r}"
        assert actual == expected, msg



"""
Application Kinds of Salt apps.
These are used to indicate what kind of Application is using RAET
"""

from collections import namedtuple

from salt.utils.odict import OrderedDict

# Python equivalent of an enum
APPL_KINDS = OrderedDict([("master", 0), ("minion", 1), ("syndic", 2), ("caller", 3)])
APPL_KIND_NAMES = OrderedDict(
    (v, k) for k, v in list(APPL_KINDS.items())
)  # inverse map
ApplKind = namedtuple("ApplKind", list(APPL_KINDS.keys()))
applKinds = ApplKind(**APPL_KINDS)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.logrotate
"""

import pytest

import salt.modules.logrotate as logrotate
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def PARSE_CONF():
    return {
        "include files": {"rsyslog": ["/var/log/syslog"]},
        "rotate": 1,
        "/var/log/wtmp": {"rotate": 1},
    }


@pytest.fixture
def configure_loader_modules():
    return {logrotate: {}}


# 'show_conf' function tests: 1


def test_show_conf():
    """
    Test if it show parsed configuration
    """
    with patch("salt.modules.logrotate._parse_conf", MagicMock(return_value=True)):
        assert logrotate.show_conf()


# 'set_' function tests: 4


def test_set(PARSE_CONF):
    """
    Test if it set a new value for a specific configuration line
    """
    with patch(
        "salt.modules.logrotate._parse_conf", MagicMock(return_value=PARSE_CONF)
    ), patch.dict(logrotate.__salt__, {"file.replace": MagicMock(return_value=True)}):
        assert logrotate.set_("rotate", "2")


def test_set_failed(PARSE_CONF):
    """
    Test if it fails to set a new value for a specific configuration line
    """
    with patch(
        "salt.modules.logrotate._parse_conf", MagicMock(return_value=PARSE_CONF)
    ):
        kwargs = {"key": "/var/log/wtmp", "value": 2}
        pytest.raises(SaltInvocationError, logrotate.set_, **kwargs)


def test_set_setting(PARSE_CONF):
    """
    Test if it set a new value for a specific configuration line
    """
    with patch.dict(
        logrotate.__salt__, {"file.replace": MagicMock(return_value=True)}
    ), patch("salt.modules.logrotate._parse_conf", MagicMock(return_value=PARSE_CONF)):
        assert logrotate.set_("/var/log/wtmp", "rotate", "2")


def test_set_setting_failed(PARSE_CONF):
    """
    Test if it fails to set a new value for a specific configuration line
    """
    with patch(
        "salt.modules.logrotate._parse_conf", MagicMock(return_value=PARSE_CONF)
    ):
        kwargs = {"key": "rotate", "value": "/var/log/wtmp", "setting": "2"}
        pytest.raises(SaltInvocationError, logrotate.set_, **kwargs)


def test_get(PARSE_CONF):
    """
    Test if get a value for a specific configuration line
    """
    with patch(
        "salt.modules.logrotate._parse_conf", MagicMock(return_value=PARSE_CONF)
    ):
        # A single key returns the right value
        assert logrotate.get("rotate") == 1

        # A single key returns the wrong value
        assert logrotate.get("rotate") != 2

        # A single key returns the right stanza value
        assert logrotate.get("/var/log/wtmp", "rotate") == 1

        # A single key returns the wrong stanza value
        assert logrotate.get("/var/log/wtmp", "rotate") != 2

        # Ensure we're logging the message as debug not warn
        with patch.object(logrotate, "_LOG") as log_mock:
            res = logrotate.get("/var/log/utmp", "rotate")
            assert log_mock.debug.called
            assert not log_mock.warn.called



"""
Module for managing locales on POSIX-like systems.
"""

import logging
import os
import re
import subprocess

import salt.utils.locales
import salt.utils.path
import salt.utils.platform
import salt.utils.systemd
from salt.exceptions import CommandExecutionError

try:
    import dbus
except ImportError:
    dbus = None


log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "locale"


def __virtual__():
    """
    Exclude Windows OS.
    """
    if salt.utils.platform.is_windows():
        return False, "Cannot load locale module: windows platforms are unsupported"

    return __virtualname__


def _parse_dbus_locale():
    """
    Get the 'System Locale' parameters from dbus
    """
    bus = dbus.SystemBus()
    localed = bus.get_object("org.freedesktop.locale1", "/org/freedesktop/locale1")
    properties = dbus.Interface(localed, "org.freedesktop.DBus.Properties")
    system_locale = properties.Get("org.freedesktop.locale1", "Locale")

    ret = {}
    for env_var in system_locale:
        env_var = str(env_var)
        match = re.match(r"^([A-Z_]+)=(.*)$", env_var)
        if match:
            ret[match.group(1)] = match.group(2).replace('"', "")
        else:
            log.error(
                'Odd locale parameter "%s" detected in dbus locale '
                "output. This should not happen. You should "
                "probably investigate what caused this.",
                env_var,
            )

    return ret


def _localectl_status():
    """
    Parse localectl status into a dict.
    :return: dict
    """
    if salt.utils.path.which("localectl") is None:
        raise CommandExecutionError('Unable to find "localectl"')
    else:
        proc = subprocess.run(["localectl"], check=False, capture_output=True)
        if b"Failed to connect to bus: No such file or directory" in proc.stderr:
            raise CommandExecutionError('Command "localectl" is in a degraded state.')

    ret = {}
    locale_ctl_out = (__salt__["cmd.run"]("localectl status") or "").strip()
    ctl_key = None
    for line in locale_ctl_out.splitlines():
        if ": " in line:  # Keys are separate with ":" and a space (!).
            ctl_key, ctl_data = line.split(": ")
            ctl_key = ctl_key.strip().lower().replace(" ", "_")
        else:
            ctl_data = line.strip()
        if not ctl_data:
            continue
        if ctl_key:
            if "=" in ctl_data:
                loc_set = ctl_data.split("=")
                if len(loc_set) == 2:
                    if ctl_key not in ret:
                        ret[ctl_key] = {}
                    ret[ctl_key][loc_set[0]] = loc_set[1]
            else:
                ret[ctl_key] = {"data": None if ctl_data == "n/a" else ctl_data}
    if not ret:
        log.debug(
            "Unable to find any locale information inside the following data:\n%s",
            locale_ctl_out,
        )
        raise CommandExecutionError('Unable to parse result of "localectl"')

    return ret


def _localectl_set(locale=""):
    """
    Use systemd's localectl command to set the LANG locale parameter, making
    sure not to trample on other params that have been set.
    """
    locale_params = (
        _parse_dbus_locale()
        if dbus is not None
        else _localectl_status().get("system_locale", {})
    )
    locale_params["LANG"] = str(locale)
    args = " ".join([f'{k}="{v}"' for k, v in locale_params.items() if v is not None])
    return not __salt__["cmd.retcode"](
        f"localectl set-locale {args}", python_shell=False
    )


def list_avail():
    """
    Lists available (compiled) locales

    CLI Example:

    .. code-block:: bash

        salt '*' locale.list_avail
    """
    return __salt__["cmd.run"]("locale -a").split("\n")


def get_locale():
    """
    Get the current system locale

    CLI Example:

    .. code-block:: bash

        salt '*' locale.get_locale
    """
    ret = ""
    lc_ctl = salt.utils.systemd.booted(__context__)
    # localectl on SLE12 is installed but the integration is still broken in latest SP3 due to
    # config is rewritten by by many %post installation hooks in the older packages.
    # If you use it -- you will break your config. This is not the case in SLE15 anymore.
    if lc_ctl and not (
        __grains__["os_family"] in ["Suse"] and __grains__["osmajorrelease"] in [12]
    ):
        ret = (
            _parse_dbus_locale()
            if dbus is not None
            else _localectl_status()["system_locale"]
        ).get("LANG", "")
    else:
        if "Suse" in __grains__["os_family"]:
            cmd = 'grep "^RC_LANG" /etc/sysconfig/language'
        elif "RedHat" in __grains__["os_family"]:
            cmd = 'grep "^LANG=" /etc/sysconfig/i18n'
        elif "Debian" in __grains__["os_family"]:
            # this block only applies to Debian without systemd
            cmd = 'grep "^LANG=" /etc/default/locale'
        elif "Gentoo" in __grains__["os_family"]:
            cmd = "eselect --brief locale show"
            return __salt__["cmd.run"](cmd).strip()
        elif "Solaris" in __grains__["os_family"]:
            cmd = 'grep "^LANG=" /etc/default/init'
        else:  # don't waste time on a failing cmd.run
            raise CommandExecutionError(
                'Error: "{}" is unsupported!'.format(__grains__["oscodename"])
            )

        if cmd:
            try:
                ret = __salt__["cmd.run"](cmd).split("=")[1].replace('"', "")
            except IndexError as err:
                log.error('Error occurred while running "%s": %s', cmd, err)

    return ret


def set_locale(locale):
    """
    Sets the current system locale

    CLI Example:

    .. code-block:: bash

        salt '*' locale.set_locale 'en_US.UTF-8'
    """
    lc_ctl = salt.utils.systemd.booted(__context__)
    # localectl on SLE12 is installed but the integration is broken -- config is rewritten by YaST2
    if lc_ctl and not (
        __grains__["os_family"] in ["Suse"] and __grains__["osmajorrelease"] in [12]
    ):
        return _localectl_set(locale)

    if "Suse" in __grains__["os_family"]:
        # this block applies to all SUSE systems - also with systemd
        if not __salt__["file.file_exists"]("/etc/sysconfig/language"):
            __salt__["file.touch"]("/etc/sysconfig/language")
        __salt__["file.replace"](
            "/etc/sysconfig/language",
            "^RC_LANG=.*",
            f'RC_LANG="{locale}"',
            append_if_not_found=True,
        )
    elif "RedHat" in __grains__["os_family"]:
        if not __salt__["file.file_exists"]("/etc/sysconfig/i18n"):
            __salt__["file.touch"]("/etc/sysconfig/i18n")
        __salt__["file.replace"](
            "/etc/sysconfig/i18n",
            "^LANG=.*",
            f'LANG="{locale}"',
            append_if_not_found=True,
        )
    elif "Debian" in __grains__["os_family"]:
        # this block only applies to Debian without systemd
        update_locale = salt.utils.path.which("update-locale")
        if update_locale is None:
            raise CommandExecutionError(
                'Cannot set locale: "update-locale" was not found.'
            )
        __salt__["cmd.run"](update_locale)  # (re)generate /etc/default/locale
        __salt__["file.replace"](
            "/etc/default/locale",
            "^LANG=.*",
            f'LANG="{locale}"',
            append_if_not_found=True,
        )
    elif "Gentoo" in __grains__["os_family"]:
        cmd = f"eselect --brief locale set {locale}"
        return __salt__["cmd.retcode"](cmd, python_shell=False) == 0
    elif "Solaris" in __grains__["os_family"]:
        if locale not in __salt__["locale.list_avail"]():
            return False
        __salt__["file.replace"](
            "/etc/default/init",
            "^LANG=.*",
            f'LANG="{locale}"',
            append_if_not_found=True,
        )
    else:
        raise CommandExecutionError("Error: Unsupported platform!")

    return True


def avail(locale):
    """
    Check if a locale is available.

    .. versionadded:: 2014.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' locale.avail 'en_US.UTF-8'
    """
    try:
        normalized_locale = salt.utils.locales.normalize_locale(locale)
    except IndexError:
        log.error('Unable to validate locale "%s"', locale)
        return False
    avail_locales = __salt__["locale.list_avail"]()
    locale_exists = next(
        (
            True
            for x in avail_locales
            if salt.utils.locales.normalize_locale(x.strip()) == normalized_locale
        ),
        False,
    )
    return locale_exists


def gen_locale(locale, **kwargs):
    """
    Generate a locale. Options:

    .. versionadded:: 2014.7.0

    :param locale: Any locale listed in /usr/share/i18n/locales or
        /usr/share/i18n/SUPPORTED for Debian and Gentoo based distributions,
        which require the charmap to be specified as part of the locale
        when generating it.

    verbose
        Show extra warnings about errors that are normally ignored.

    CLI Example:

    .. code-block:: bash

        salt '*' locale.gen_locale en_US.UTF-8
        salt '*' locale.gen_locale 'en_IE.UTF-8 UTF-8'    # Debian/Gentoo only
    """
    on_debian = __grains__.get("os") == "Debian"
    on_ubuntu = __grains__.get("os") == "Ubuntu"
    on_gentoo = __grains__.get("os_family") == "Gentoo"
    on_suse = __grains__.get("os_family") == "Suse"
    on_solaris = __grains__.get("os_family") == "Solaris"

    if on_solaris:  # all locales are pre-generated
        return locale in __salt__["locale.list_avail"]()

    locale_info = salt.utils.locales.split_locale(locale)
    locale_search_str = "{}_{}".format(
        locale_info["language"], locale_info["territory"]
    )

    # if the charmap has not been supplied, normalize by appening it
    if not locale_info["charmap"] and not on_ubuntu:
        locale_info["charmap"] = locale_info["codeset"]
        locale = salt.utils.locales.join_locale(locale_info)

    if on_debian or on_gentoo:  # file-based search
        search = "/usr/share/i18n/SUPPORTED"
        valid = __salt__["file.search"](search, f"^{locale}$", flags=re.MULTILINE)
    else:  # directory-based search
        if on_suse:
            search = "/usr/share/locale"
        else:
            search = "/usr/share/i18n/locales"

        try:
            valid = locale_search_str in os.listdir(search)
        except OSError as ex:
            log.error(ex)
            raise CommandExecutionError(f'Locale "{locale}" is not available.')

    if not valid:
        log.error('The provided locale "%s" is not found in %s', locale, search)
        return False

    if os.path.exists("/etc/locale.gen"):
        __salt__["file.replace"](
            "/etc/locale.gen",
            rf"^\s*#\s*{locale}\s*$",
            f"{locale}\n",
            append_if_not_found=True,
        )
    elif on_ubuntu:
        __salt__["file.touch"](
            "/var/lib/locales/supported.d/{}".format(locale_info["language"])
        )
        __salt__["file.replace"](
            "/var/lib/locales/supported.d/{}".format(locale_info["language"]),
            locale,
            locale,
            append_if_not_found=True,
        )

    if salt.utils.path.which("locale-gen"):
        cmd = ["locale-gen"]
        if on_gentoo:
            cmd.append("--generate")
        if on_ubuntu:
            cmd.append(salt.utils.locales.normalize_locale(locale))
        else:
            cmd.append(locale)
    elif salt.utils.path.which("localedef"):
        cmd = [
            "localedef",
            "--force",
            "-i",
            locale_search_str,
            "-f",
            locale_info["codeset"],
            "{}.{}".format(locale_search_str, locale_info["codeset"]),
            kwargs.get("verbose", False) and "--verbose" or "--quiet",
        ]
    else:
        raise CommandExecutionError(
            'Command "locale-gen" or "localedef" was not found on this system.'
        )

    res = __salt__["cmd.run_all"](cmd)
    if res["retcode"]:
        log.error(res["stderr"])

    if kwargs.get("verbose"):
        return res
    else:
        return res["retcode"] == 0



import os

import salt.utils.package


def test_pkg_type():
    ret = salt.utils.package.pkg_type()
    if os.environ.get("ONEDIR_TESTRUN", "0") == "0":
        assert ret == "pip"
    else:
        assert ret == "onedir"



import pytest

import salt.modules.win_dism as dism
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {dism: {}}


def test_add_capability():
    """
    Test installing a capability with DISM
    """
    mock = MagicMock()
    with patch.dict(dism.__salt__, {"cmd.run_all": mock}):
        with patch.dict(dism.__grains__, {"osversion": 10}):
            dism.add_capability("test")
            mock.assert_called_once_with(
                [
                    dism.bin_dism,
                    "/Quiet",
                    "/Online",
                    "/Add-Capability",
                    "/CapabilityName:test",
                    "/NoRestart",
                ]
            )


def test_add_capability_with_extras():
    """
    Test installing a capability with DISM
    """
    mock = MagicMock()
    with patch.dict(dism.__salt__, {"cmd.run_all": mock}):
        with patch.dict(dism.__grains__, {"osversion": 10}):
            dism.add_capability("test", "life", True)
            mock.assert_called_once_with(
                [
                    dism.bin_dism,
                    "/Quiet",
                    "/Online",
                    "/Add-Capability",
                    "/CapabilityName:test",
                    "/Source:life",
                    "/LimitAccess",
                    "/NoRestart",
                ]
            )


def test_remove_capability():
    """
    Test uninstalling a capability with DISM
    """
    mock = MagicMock()
    with patch.dict(dism.__salt__, {"cmd.run_all": mock}):
        with patch.dict(dism.__grains__, {"osversion": 10}):
            dism.remove_capability("test")
            mock.assert_called_once_with(
                [
                    dism.bin_dism,
                    "/Quiet",
                    "/Online",
                    "/Remove-Capability",
                    "/CapabilityName:test",
                    "/NoRestart",
                ]
            )


def test_get_capabilities():
    """
    Test getting all the capabilities
    """
    capabilties = (
        "Capability Identity : Capa1\r\n State : Installed\r\n"
        "Capability Identity : Capa2\r\n State : Disabled\r\n"
    )

    mock = MagicMock(return_value=capabilties)
    with patch.dict(dism.__salt__, {"cmd.run": mock}):
        with patch.dict(dism.__grains__, {"osversion": 10}):
            out = dism.get_capabilities()
            mock.assert_called_once_with(
                [dism.bin_dism, "/English", "/Online", "/Get-Capabilities"]
            )
            assert out == ["Capa1", "Capa2"]


def test_installed_capabilities():
    """
    Test getting all the installed capabilities
    """
    capabilties = (
        "Capability Identity : Capa1\r\n State : Installed\r\n"
        "Capability Identity : Capa2\r\n State : Disabled\r\n"
    )

    mock = MagicMock(return_value=capabilties)
    with patch.dict(dism.__salt__, {"cmd.run": mock}):
        with patch.dict(dism.__grains__, {"osversion": 10}):
            out = dism.installed_capabilities()
            mock.assert_called_once_with(
                [dism.bin_dism, "/English", "/Online", "/Get-Capabilities"]
            )
            assert out == ["Capa1"]


def test_available_capabilities():
    """
    Test getting all the available capabilities
    """
    capabilties = (
        "Capability Identity : Capa1\r\n State : Installed\r\n"
        "Capability Identity : Capa2\r\n State : Not Present\r\n"
    )

    mock = MagicMock(return_value=capabilties)
    with patch.dict(dism.__salt__, {"cmd.run": mock}):
        with patch.dict(dism.__grains__, {"osversion": 10}):
            out = dism.available_capabilities()
            mock.assert_called_once_with(
                [dism.bin_dism, "/English", "/Online", "/Get-Capabilities"]
            )
            assert out == ["Capa2"]


def test_add_feature():
    """
    Test installing a feature with DISM
    """
    mock = MagicMock()
    with patch.dict(dism.__salt__, {"cmd.run_all": mock}):
        dism.add_feature("test")
        mock.assert_called_once_with(
            [
                dism.bin_dism,
                "/Quiet",
                "/Online",
                "/Enable-Feature",
                "/FeatureName:test",
                "/NoRestart",
            ]
        )


def test_add_feature_with_extras():
    """
    Test installing a feature with DISM
    """
    mock = MagicMock()
    with patch.dict(dism.__salt__, {"cmd.run_all": mock}):
        dism.add_feature("sponge", "bob", "C:\\temp", True, True)
        mock.assert_called_once_with(
            [
                dism.bin_dism,
                "/Quiet",
                "/Online",
                "/Enable-Feature",
                "/FeatureName:sponge",
                "/PackageName:bob",
                "/Source:C:\\temp",
                "/LimitAccess",
                "/All",
                "/NoRestart",
            ]
        )


def test_remove_feature():
    """
    Test uninstalling a capability with DISM
    """
    mock = MagicMock()
    with patch.dict(dism.__salt__, {"cmd.run_all": mock}):
        dism.remove_feature("test")
        mock.assert_called_once_with(
            [
                dism.bin_dism,
                "/Quiet",
                "/Online",
                "/Disable-Feature",
                "/FeatureName:test",
                "/NoRestart",
            ]
        )


def test_remove_feature_with_extras():
    """
    Test uninstalling a capability with DISM
    """
    mock = MagicMock()
    with patch.dict(dism.__salt__, {"cmd.run_all": mock}):
        dism.remove_feature("sponge", True)
        mock.assert_called_once_with(
            [
                dism.bin_dism,
                "/Quiet",
                "/Online",
                "/Disable-Feature",
                "/FeatureName:sponge",
                "/Remove",
                "/NoRestart",
            ]
        )


def test_get_features():
    """
    Test getting all the features
    """
    features = (
        "Feature Name : Capa1\r\n State : Enabled\r\n"
        "Feature Name : Capa2\r\n State : Disabled\r\n"
    )

    mock = MagicMock(return_value=features)
    with patch.dict(dism.__salt__, {"cmd.run": mock}):
        out = dism.get_features()
        mock.assert_called_once_with(
            [dism.bin_dism, "/English", "/Online", "/Get-Features"]
        )
        assert out == ["Capa1", "Capa2"]


def test_installed_features():
    """
    Test getting all the installed features
    """
    features = (
        "Feature Name : Capa1\r\n State : Enabled\r\n"
        "Feature Name : Capa2\r\n State : Disabled\r\n"
    )

    mock = MagicMock(return_value=features)
    with patch.dict(dism.__salt__, {"cmd.run": mock}):
        out = dism.installed_features()
        mock.assert_called_once_with(
            [dism.bin_dism, "/English", "/Online", "/Get-Features"]
        )
        assert out == ["Capa1"]


def test_available_features():
    """
    Test getting all the available features
    """
    features = (
        "Feature Name : Capa1\r\n State : Enabled\r\n"
        "Feature Name : Capa2\r\n State : Disabled\r\n"
    )

    mock = MagicMock(return_value=features)
    with patch.dict(dism.__salt__, {"cmd.run": mock}):
        out = dism.available_features()
        mock.assert_called_once_with(
            [dism.bin_dism, "/English", "/Online", "/Get-Features"]
        )
        assert out == ["Capa2"]


def test_add_package():
    """
    Test installing a package with DISM
    """
    mock = MagicMock()
    with patch.dict(dism.__salt__, {"cmd.run_all": mock}):
        dism.add_package("test")
        mock.assert_called_once_with(
            [
                dism.bin_dism,
                "/Quiet",
                "/Online",
                "/Add-Package",
                "/PackagePath:test",
                "/NoRestart",
            ]
        )


def test_add_package_with_extras():
    """
    Test installing a package with DISM
    """
    mock = MagicMock()
    with patch.dict(dism.__salt__, {"cmd.run_all": mock}):
        dism.add_package("sponge", True, True)
        mock.assert_called_once_with(
            [
                dism.bin_dism,
                "/Quiet",
                "/Online",
                "/Add-Package",
                "/PackagePath:sponge",
                "/IgnoreCheck",
                "/PreventPending",
                "/NoRestart",
            ]
        )


def test_remove_package():
    """
    Test uninstalling a package with DISM
    """
    mock = MagicMock()
    with patch.dict(dism.__salt__, {"cmd.run_all": mock}):
        dism.remove_package("test")
        mock.assert_called_once_with(
            [
                dism.bin_dism,
                "/Quiet",
                "/Online",
                "/Remove-Package",
                "/NoRestart",
                "/PackagePath:test",
            ]
        )


def test_remove_kb():
    """
    Test uninstalling a KB with DISM
    """
    pkg_name = "Package_for_KB1002345~31bf3856ad364e35~amd64~~22000.345.1.1"
    mock_search = MagicMock(return_value=[pkg_name])
    mock_remove = MagicMock()
    with patch("salt.modules.win_dism.installed_packages", mock_search):
        with patch("salt.modules.win_dism.remove_package", mock_remove):
            dism.remove_kb("KB1002345")
            mock_remove.assert_called_once_with(
                package=pkg_name,
                image=None,
                restart=False,
            )


def test_remove_kb_number():
    """
    Test uninstalling a KB with DISM with just the KB number
    """
    pkg_name = "Package_for_KB1002345~31bf3856ad364e35~amd64~~22000.345.1.1"
    mock_search = MagicMock(return_value=[pkg_name])
    mock_remove = MagicMock()
    with patch("salt.modules.win_dism.installed_packages", mock_search):
        with patch("salt.modules.win_dism.remove_package", mock_remove):
            dism.remove_kb("1002345")
            mock_remove.assert_called_once_with(
                package=pkg_name,
                image=None,
                restart=False,
            )


def test_remove_kb_not_found():
    pkg_name = "Package_for_KB1002345~31bf3856ad364e35~amd64~~22000.345.1.1"
    mock_search = MagicMock(return_value=[pkg_name])
    with patch("salt.modules.win_dism.installed_packages", mock_search):
        with pytest.raises(CommandExecutionError) as err:
            dism.remove_kb("1001111")
        assert str(err.value) == "1001111 not installed"


def test_installed_packages():
    """
    Test getting all the installed features
    """
    features = (
        "Package Identity : Capa1\r\n State : Installed\r\n"
        "Package Identity : Capa2\r\n State : Installed\r\n"
    )

    mock = MagicMock(return_value=features)
    with patch.dict(dism.__salt__, {"cmd.run": mock}):
        out = dism.installed_packages()
        mock.assert_called_once_with(
            [dism.bin_dism, "/English", "/Online", "/Get-Packages"]
        )
        assert out == ["Capa1", "Capa2"]



"""
    tests.support.events
    ~~~~~~~~~~~~~~~~~~~~
"""

import multiprocessing
import os
import time
from contextlib import contextmanager

import salt.utils.event
from salt.utils.process import Process, clean_proc


@contextmanager
def eventpublisher_process(sock_dir):
    opts = {
        "sock_dir": sock_dir,
        "interface": "127.0.0.1",
        "publish_port": 4506,
        "ipv6": None,
        "zmq_filtering": None,
    }
    ipc_publisher = salt.transport.publish_server(
        opts,
        pub_path=os.path.join(opts["sock_dir"], "master_event_pub.ipc"),
        pull_path=os.path.join(opts["sock_dir"], "master_event_pull.ipc"),
        transport="tcp",
    )
    proc = Process(
        target=ipc_publisher.publish_daemon,
        args=[
            ipc_publisher.publish_payload,
        ],
    )
    # proc = salt.utils.event.EventPublisher({"sock_dir": sock_dir})
    proc.start()
    try:
        if os.environ.get("TRAVIS_PYTHON_VERSION", None) is not None:
            # Travis is slow
            time.sleep(10)
        else:
            time.sleep(8)
        yield
    finally:
        clean_proc(proc)


class EventSender(multiprocessing.Process):
    def __init__(self, data, tag, wait, sock_dir):
        super().__init__()
        self.data = data
        self.tag = tag
        self.wait = wait
        self.sock_dir = sock_dir

    def run(self):
        with salt.utils.event.MasterEvent(self.sock_dir, listen=False) as me:
            time.sleep(self.wait)
            me.fire_event(self.data, self.tag)
            # Wait a few seconds before tearing down the zmq context
            if os.environ.get("TRAVIS_PYTHON_VERSION", None) is not None:
                # Travis is slow
                time.sleep(10)
            else:
                time.sleep(2)


@contextmanager
def eventsender_process(data, tag, sock_dir, wait=0):
    proc = EventSender(data, tag, wait, sock_dir=sock_dir)
    proc.start()
    try:
        yield
    finally:
        clean_proc(proc)



import logging
import re

import requests
from requests.packages.urllib3.util.ssl_ import create_urllib3_context

import salt.exceptions
import salt.utils.vault.leases as leases
from salt.utils.vault.exceptions import (
    VaultAuthExpired,
    VaultInvocationError,
    VaultNotFoundError,
    VaultPermissionDeniedError,
    VaultPreconditionFailedError,
    VaultServerError,
    VaultUnavailableError,
    VaultUnsupportedOperationError,
    VaultUnwrapException,
)

log = logging.getLogger(__name__)
logging.getLogger("requests").setLevel(logging.WARNING)

# This list is not complete at all, but contains
# the most important paths.
VAULT_UNAUTHD_PATHS = (
    "sys/wrapping/lookup",
    "sys/internal/ui/mounts",
    "sys/internal/ui/namespaces",
    "sys/seal-status",
    "sys/health",
)


def _get_expected_creation_path(secret_type, config=None):
    if secret_type == "token":
        return r"auth/token/create(/[^/]+)?"

    if secret_type == "secret_id":
        if config is not None:
            return r"auth/{}/role/{}/secret\-id".format(
                re.escape(config["auth"]["approle_mount"]),
                re.escape(config["auth"]["approle_name"]),
            )
        return r"auth/[^/]+/role/[^/]+/secret\-id"

    if secret_type == "role_id":
        if config is not None:
            return r"auth/{}/role/{}/role\-id".format(
                re.escape(config["auth"]["approle_mount"]),
                re.escape(config["auth"]["approle_name"]),
            )
        return r"auth/[^/]+/role/[^/]+/role\-id"

    raise salt.exceptions.SaltInvocationError(
        f"secret_type must be one of token, secret_id, role_id, got `{secret_type}`."
    )


class VaultClient:
    """
    Unauthenticated client for the Vault API.
    Base class for authenticated client.
    """

    def __init__(self, url, namespace=None, verify=None, session=None):
        self.url = url
        self.namespace = namespace
        self.verify = verify

        ca_cert = None
        try:
            if verify.startswith("-----BEGIN CERTIFICATE"):
                ca_cert = verify
                verify = None
        except AttributeError:
            pass

        # Keep the actual requests parameter separate from the client config
        # to reduce complexity in config validation.
        self._requests_verify = verify
        if session is None:
            session = requests.Session()
            if ca_cert:
                adapter = CACertHTTPSAdapter(ca_cert)
                session.mount(url, adapter)
        self.session = session

    def delete(self, endpoint, wrap=False, raise_error=True, add_headers=None):
        """
        Wrapper for client.request("DELETE", ...)
        """
        return self.request(
            "DELETE",
            endpoint,
            wrap=wrap,
            raise_error=raise_error,
            add_headers=add_headers,
        )

    def get(self, endpoint, wrap=False, raise_error=True, add_headers=None):
        """
        Wrapper for client.request("GET", ...)
        """
        return self.request(
            "GET", endpoint, wrap=wrap, raise_error=raise_error, add_headers=add_headers
        )

    def list(self, endpoint, wrap=False, raise_error=True, add_headers=None):
        """
        Wrapper for client.request("LIST", ...)
        TODO: configuration to enable GET requests with query parameters for LIST?
        """
        return self.request(
            "LIST",
            endpoint,
            wrap=wrap,
            raise_error=raise_error,
            add_headers=add_headers,
        )

    def post(
        self, endpoint, payload=None, wrap=False, raise_error=True, add_headers=None
    ):
        """
        Wrapper for client.request("POST", ...)
        Vault considers POST and PUT to be synonymous.
        """
        return self.request(
            "POST",
            endpoint,
            payload=payload,
            wrap=wrap,
            raise_error=raise_error,
            add_headers=add_headers,
        )

    def patch(self, endpoint, payload, wrap=False, raise_error=True, add_headers=None):
        """
        Wrapper for client.request("PATCH", ...)
        """
        return self.request(
            "PATCH",
            endpoint,
            payload=payload,
            wrap=wrap,
            raise_error=raise_error,
            add_headers=add_headers,
        )

    def request(
        self,
        method,
        endpoint,
        payload=None,
        wrap=False,
        raise_error=True,
        add_headers=None,
        **kwargs,
    ):
        """
        Issue a request against the Vault API.
        Returns boolean when no data was returned, otherwise the decoded json data
        or a VaultWrappedResponse object if wrapping was requested.
        """
        res = self.request_raw(
            method,
            endpoint,
            payload=payload,
            wrap=wrap,
            add_headers=add_headers,
            **kwargs,
        )
        if res.status_code == 204:
            return True
        data = res.json()
        if not res.ok:
            if raise_error:
                self._raise_status(res)
            return data
        if wrap:
            return leases.VaultWrappedResponse(**data["wrap_info"])
        return data

    def request_raw(
        self, method, endpoint, payload=None, wrap=False, add_headers=None, **kwargs
    ):
        """
        Issue a request against the Vault API. Returns the raw response object.
        """
        url = self._get_url(endpoint)
        headers = self._get_headers(wrap)
        try:
            headers.update(add_headers)
        except TypeError:
            pass
        res = self.session.request(
            method,
            url,
            headers=headers,
            json=payload,
            verify=self._requests_verify,
            **kwargs,
        )
        return res

    def unwrap(self, wrapped, expected_creation_path=None):
        """
        Unwraps the data associated with a wrapping token.

        wrapped
            Wrapping token to unwrap

        expected_creation_path
            Regex expression or list of expressions that should fully match the
            wrapping token creation path. At least one match is required.
            Defaults to None, which skips the check.

            .. note::
                This check prevents tampering with wrapping tokens, which are
                valid for one request only. Usually, if an attacker sniffs a wrapping
                token, there will be two unwrapping requests, causing an audit warning.
                If the attacker can issue a new wrapping token and insert it into the
                response instead, this warning would be silenced. Assuming they do not
                possess the permissions to issue a wrapping token from the correct
                endpoint, checking the creation path makes this kind of attack obvious.
        """
        if expected_creation_path:
            wrap_info = self.wrap_info(wrapped)
            if not isinstance(expected_creation_path, list):
                expected_creation_path = [expected_creation_path]
            if not any(
                re.fullmatch(p, wrap_info["creation_path"])
                for p in expected_creation_path
            ):
                raise VaultUnwrapException(
                    actual=wrap_info["creation_path"],
                    expected=expected_creation_path,
                    url=self.url,
                    namespace=self.namespace,
                    verify=self.verify,
                )
        endpoint = "sys/wrapping/unwrap"
        headers = self._get_headers()
        payload = {}
        if "X-Vault-Token" not in headers:
            headers["X-Vault-Token"] = str(wrapped)
        else:
            payload["token"] = str(wrapped)
        return self.post(endpoint=endpoint, add_headers=headers, payload=payload)

    def wrap_info(self, wrapped):
        """
        Lookup wrapping token meta information.
        """
        endpoint = "sys/wrapping/lookup"
        add_headers = {"X-Vault-Token": str(wrapped)}
        return self.post(endpoint, wrap=False, add_headers=add_headers)["data"]

    def token_lookup(self, token=None, accessor=None, raw=False):
        """
        Lookup token meta information.

        token
            The token to look up or to use to look up the accessor.
            Required.

        accessor
            The accessor to use to query the token meta information.

        raw
            Return the raw response object instead of response data.
            Also disables status code checking.
        """
        endpoint = "auth/token/lookup-self"
        method = "GET"
        payload = {}
        if token is None:
            raise VaultInvocationError(
                "Unauthenticated VaultClient needs a token to lookup."
            )
        add_headers = {"X-Vault-Token": token}

        if accessor is not None:
            endpoint = "auth/token/lookup-accessor"
            payload["accessor"] = accessor

        res = self.request_raw(
            method, endpoint, payload=payload, wrap=False, add_headers=add_headers
        )
        if raw:
            return res
        self._raise_status(res)
        return res.json()["data"]

    def token_valid(self, valid_for=0, remote=True):  # pylint: disable=unused-argument
        return False

    def get_config(self):
        """
        Returns Vault server configuration used by this client.
        """
        return {
            "url": self.url,
            "namespace": self.namespace,
            "verify": self.verify,
        }

    def _get_url(self, endpoint):
        endpoint = endpoint.strip("/")
        return f"{self.url}/v1/{endpoint}"

    def _get_headers(self, wrap=False):
        headers = {"Content-Type": "application/json", "X-Vault-Request": "true"}
        if self.namespace is not None:
            headers["X-Vault-Namespace"] = self.namespace
        if wrap:
            headers["X-Vault-Wrap-TTL"] = str(wrap)
        return headers

    def _raise_status(self, res):
        errors = ", ".join(res.json().get("errors", []))
        if res.status_code == 400:
            raise VaultInvocationError(errors)
        if res.status_code == 403:
            raise VaultPermissionDeniedError(errors)
        if res.status_code == 404:
            raise VaultNotFoundError(errors)
        if res.status_code == 405:
            raise VaultUnsupportedOperationError(errors)
        if res.status_code == 412:
            raise VaultPreconditionFailedError(errors)
        if res.status_code in [500, 502]:
            raise VaultServerError(errors)
        if res.status_code == 503:
            raise VaultUnavailableError(errors)
        res.raise_for_status()


class AuthenticatedVaultClient(VaultClient):
    """
    Authenticated client for the Vault API.
    This should be used for most operations.
    """

    auth = None

    def __init__(self, auth, url, **kwargs):
        self.auth = auth
        super().__init__(url, **kwargs)

    def token_valid(self, valid_for=0, remote=True):
        """
        Check whether this client's authentication information is
        still valid.

        remote
            Check with the remote Vault server as well. This consumes
            a token use. Defaults to true.
        """
        if not self.auth.is_valid(valid_for):
            return False
        if not remote:
            return True
        try:
            res = self.token_lookup(raw=True)
            if res.status_code != 200:
                return False
            return True
        except Exception as err:  # pylint: disable=broad-except
            raise salt.exceptions.CommandExecutionError(
                "Error while looking up self token."
            ) from err

    def token_lookup(self, token=None, accessor=None, raw=False):
        """
        Lookup token meta information.

        token
            The token to look up. If neither token nor accessor
            are specified, looks up the current token in use by
            this client.

        accessor
            The accessor of the token to query the meta information for.

        raw
            Return the raw response object instead of response data.
            Also disables status code checking.
        """
        endpoint = "auth/token/lookup"
        method = "POST"
        payload = {}
        if token is None and accessor is None:
            endpoint += "-self"
            method = "GET"
        if token is not None:
            payload["token"] = token
        elif accessor is not None:
            endpoint += "-accessor"
            payload["accessor"] = accessor
        if raw:
            return self.request_raw(method, endpoint, payload=payload, wrap=False)
        return self.request(method, endpoint, payload=payload, wrap=False)["data"]

    def token_renew(self, increment=None, token=None, accessor=None):
        """
        Renew a token.

        increment
            Request the token to be valid for this amount of time from the current
            point of time onwards. Can also be used to reduce the validity period.
            The server might not honor this increment.
            Can be an integer (seconds) or a time string like ``1h``. Optional.

        token
            The token that should be renewed. Optional.
            If token and accessor are unset, renews the token currently in use
            by this client.

        accessor
            The accessor of the token that should be renewed. Optional.
        """
        endpoint = "auth/token/renew"
        payload = {}

        if token is None and accessor is None:
            if not self.auth.is_renewable():
                return False
            endpoint += "-self"

        if increment is not None:
            payload["increment"] = increment
        if token is not None:
            payload["token"] = token
        elif accessor is not None:
            endpoint += "-accessor"
            payload["accessor"] = accessor

        res = self.post(endpoint, payload=payload)

        if token is None and accessor is None:
            self.auth.update_token(res["auth"])
        return res["auth"]

    def token_revoke(self, delta=1, token=None, accessor=None):
        """
        Revoke a token by setting its TTL to 1s.

        delta
            The time in seconds to request revocation after.
            Defaults to 1s.

        token
            The token that should be revoked. Optional.
            If token and accessor are unset, revokes the token currently in use
            by this client.

        accessor
            The accessor of the token that should be revoked. Optional.
        """
        try:
            self.token_renew(increment=delta, token=token, accessor=accessor)
        except (VaultPermissionDeniedError, VaultNotFoundError, VaultAuthExpired):
            # if we're trying to revoke ourselves and this happens,
            # the token was already invalid
            if token or accessor:
                raise
            return False
        return True

    def request_raw(
        self,
        method,
        endpoint,
        payload=None,
        wrap=False,
        add_headers=None,
        is_unauthd=False,
        **kwargs,
    ):  # pylint: disable=arguments-differ
        """
        Issue an authenticated request against the Vault API. Returns the raw response object.
        """
        ret = super().request_raw(
            method,
            endpoint,
            payload=payload,
            wrap=wrap,
            add_headers=add_headers,
            **kwargs,
        )
        # tokens are used regardless of status code
        if not is_unauthd and not endpoint.startswith(VAULT_UNAUTHD_PATHS):
            self.auth.used()
        return ret

    def _get_headers(self, wrap=False):
        headers = super()._get_headers(wrap)
        headers["X-Vault-Token"] = str(self.auth.get_token())
        return headers


class CACertHTTPSAdapter(requests.sessions.HTTPAdapter):
    """
    Allows to restrict requests CA chain validation
    to a single root certificate without writing it to disk.
    """

    def __init__(self, ca_cert_data, *args, **kwargs):
        self.ca_cert_data = ca_cert_data
        super().__init__(*args, **kwargs)

    def init_poolmanager(
        self,
        connections,
        maxsize,
        block=requests.adapters.DEFAULT_POOLBLOCK,
        **pool_kwargs,
    ):
        ssl_context = create_urllib3_context()
        ssl_context.load_verify_locations(cadata=self.ca_cert_data)
        pool_kwargs["ssl_context"] = ssl_context
        return super().init_poolmanager(
            connections, maxsize, block=block, **pool_kwargs
        )



import urllib.parse

import pytest
from tornado.httpclient import HTTPError

pytestmark = [
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
]


@pytest.fixture
def auth_creds(auth_creds):
    auth_creds["eauth"] = "pam"
    return auth_creds


@pytest.mark.parametrize("service", ["chsh", "login"])
async def test_bad_pwd_pam_chsh_service(
    http_client, auth_creds, content_type_map, service
):
    """
    Test login while specifying `chsh` or `login` service with bad passwd
    This test ensures this PR is working correctly:
    https://github.com/saltstack/salt/pull/31826
    """
    auth_creds["service"] = service
    auth_creds["password"] = "wrong_password"
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/login",
            method="POST",
            body=urllib.parse.urlencode(auth_creds),
            headers={"Content-Type": content_type_map["form"]},
        )
    assert exc.value.code == 401


@pytest.mark.parametrize("service", ["chsh", "login"])
async def test_good_pwd_pam_chsh_service(
    http_client, auth_creds, content_type_map, service
):
    """
    Test login while specifying `chsh` and `login` service with good passwd
    This test ensures this PR is working correctly:
    https://github.com/saltstack/salt/pull/31826
    """
    auth_creds["service"] = service
    response = await http_client.fetch(
        "/login",
        method="POST",
        body=urllib.parse.urlencode(auth_creds),
        headers={"Content-Type": content_type_map["form"]},
    )
    assert response.code == 200
